# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
from logging import info
from tempfile import NamedTemporaryFile
from typing import Set

from portmod.cfg_protect import (
    get_changes,
    get_mergetool,
    get_redirections,
    merge_files,
    remove_redirection,
)
from portmod.config import variable_data_dir
from portmod.globals import env
from portmod.merge import global_updates
from portmod.prompt import prompt_num, prompt_options
from portmodlib.colour import bright
from portmodlib.l10n import l10n
from portmodlib.parsers.list import add_list, read_list


def apply_changes_to_file(dst: str, src: str, whitelist: Set[str]):
    whitelist_file = os.path.join(variable_data_dir(), "module-data", "file-whitelist")
    blacklist_file = os.path.join(variable_data_dir(), "module-data", "file-blacklist")
    # Display file changes to user and prompt
    original_src = src

    while True:
        output = get_changes(src, dst)
        if dst in whitelist:
            # User won't be prompted, so we should still display output, but supress it
            # unless running verbosely
            info("".join(output))
        else:
            print("".join(output))

        print()

        if dst not in whitelist and not env.INTERACTIVE:
            info(l10n("skipped-update-noninteractive", file=dst))
            break

        response = None
        if dst not in whitelist:
            options = [
                (l10n("yes-short"), l10n("apply-change")),
                (l10n("always-short"), l10n("merge-apply-always")),
                (l10n("skip-change.short"), l10n("skip-change")),
                (l10n("no-short"), l10n("merge-do-not-apply-change")),
                (l10n("never-short"), l10n("merge-apply-never")),
            ]
            mergetool = get_mergetool()
            if mergetool:
                options.append(
                    (
                        l10n("mergetool.short"),
                        l10n("mergetool", mergetool=f'"{mergetool}"'),
                    )
                )
            else:
                print(l10n("mergetool-info", var="MERGE_TOOL"))
                print()

            try:
                response = prompt_options(
                    l10n("apply-above-change-qn", file=bright(dst)), options
                )
            except EOFError:
                response = l10n("no-short")

        if response == l10n("skip-change.short"):
            break

        if dst in whitelist or response in (
            l10n("yes-short"),
            l10n("always-short"),
        ):
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            if os.path.exists(dst) or os.path.islink(dst):
                os.remove(dst)
            shutil.move(src, dst)
            # Remove the original file if src now points to a merged file
            if original_src != src:
                os.remove(original_src)

            if response == l10n("always-short"):
                add_list(whitelist_file, dst)

            remove_redirection(src, dst)
            break

        if response == l10n("mergetool.short"):
            with open(dst, "r") as file:
                contents = file.read()
            with NamedTemporaryFile(
                "w",
                delete=False,
                prefix="EDITME.",
                suffix="." + os.path.basename(dst),
            ) as tempfile:
                tempfile.write(contents)
                tempfile_path = tempfile.name
            # Merge against intermediate file
            merge_files(src, tempfile_path)
            # Don't break, as we must let the user accept the merged changes first
            src = tempfile_path

        if response in {l10n("no-short"), l10n("never-short")}:
            if response == l10n("never-short"):
                add_list(blacklist_file, dst)

            os.remove(src)
            remove_redirection(src, dst)
            break


def cfg_update(args):
    files = get_redirections()
    if not files:
        info(l10n("cfg-update-nothing-to-do"))
        return

    # Read only once rather than once per file
    whitelist = set()
    whitelist_file = os.path.join(variable_data_dir(), "module-data", "file-whitelist")
    if os.path.exists(whitelist_file):
        whitelist = set(read_list(whitelist_file))

    while files:
        # Prompt for which file to update, then apply each update in order
        for index, (dst, newfiles) in enumerate(files.items()):
            print(
                str(index + 1) + ")",
                l10n("pending-updates-for-file", file=dst, num=len(newfiles)),
            )

        index = 0
        if len(files) > 1:
            try:
                index = prompt_num(
                    l10n("update-file-prompt"), max_val=len(files), cancel=True
                )
            except EOFError:
                break

        dst, newfiles = list(files.items())[index - 1]
        for index, file in enumerate(newfiles):
            info(l10n("cfg-update-file-index", index=index + 1, total=len(newfiles)))
            apply_changes_to_file(dst, file, whitelist)

        files = get_redirections()


def add_cfg_update_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "cfg-update",
        help=l10n("cfg-update-help"),
        parents=parents,
        conflict_handler="resolve",
    )

    parser.set_defaults(func=cfg_update)


def add_module_update_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "module-update",
        help=l10n("module-update-help"),
        parents=parents,
        conflict_handler="resolve",
    )

    parser.set_defaults(func=lambda args: global_updates())



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Portmod config tests
"""

import os

import pytest

from portmod.config import get_config, get_config_value, set_config_value
from portmod.globals import env

from .env import setup_env, tear_down_env
from .merge import merge


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_profile_only_variables(setup):
    """
    Tests that sorting the config files works properly
    """
    get_config()
    with open(env.prefix().CONFIG, "w") as configfile:
        print(
            """
USE_EXPAND = "FOO"
""",
            file=configfile,
        )
    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()

    with open(env.prefix().CONFIG, "w") as configfile:
        print(
            """
ARCH = "BAR"
""",
            file=configfile,
        )

    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()

    with open(env.prefix().CONFIG, "w") as configfile:
        print(
            """
TEST_PROFILE_ONLY = "BAR"
""",
            file=configfile,
        )

    get_config.cache_clear()
    with pytest.raises(UserWarning):
        get_config()


def test_set_config_value():
    def test():
        set_config_value("foo", "1")
        assert get_config()["foo"] == "1"

    # Test with other variable
    with open(env.prefix().CONFIG, "w") as file:
        file.write('bar = "baz"')
    test()

    # Test with just comment
    with open(env.prefix().CONFIG, "w") as file:
        file.write("# Comment")
    test()

    # Test with empty file
    with open(env.prefix().CONFIG, "w"):
        pass
    test()

    # Test with nonexistant file
    os.remove(env.prefix().CONFIG)
    test()


def test_arch_version():
    merge(["modules/version"])
    get_config.cache_clear()
    assert get_config_value("ARCH_VERSION") == "1.0"



# Copyright 2024 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Backport of a typing functions to work with python 3.8
"""

from typing import Any

try:
    from typing import assert_never  # type: ignore
except ImportError:

    def assert_never(value: Any):
        raise AssertionError(f"Expected code to be unreachable, but got: {value}")



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import subprocess
import sys
from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory
from threading import Lock, Thread
from typing import AbstractSet, Iterable, List, Set, Tuple

from portmodlib.fs import is_parent
from portmodlib.l10n import l10n

from .config import get_config, get_config_value
from .globals import env
from .perms import Permissions
from .win32 import get_personal

if sys.platform == "win32":
    SBIE_LOCK = Lock()


def _required_exe(name: str) -> str:
    path = which(name)

    if not path:
        raise FileNotFoundError(l10n("missing-executable", file=name))
    return path


def resolve_paths(path: str) -> List[str]:
    """
    Returns all paths related to the symlink chain at paths
    If no symlinks exist, just returns path
    """
    paths = []
    prefix = ""
    for part in Path(path).parts:
        if os.path.islink(os.path.join(prefix, part)):
            last = os.path.join(prefix, part)
            paths.append(last)
            while os.path.islink(last):
                linkpath = os.readlink(last)
                if os.path.isabs(linkpath):
                    last = linkpath
                else:
                    last = os.path.normpath(
                        os.path.join(os.path.dirname(last), linkpath)
                    )
                paths.append(last)
        prefix = os.path.realpath(os.path.join(prefix, part))

    return paths


def parents(paths: Set[str]) -> Set[str]:
    results = set()
    for path in paths:
        for ppath in Path(path).parents:
            results.add(str(ppath))
    return results


def _dedup_subpaths(paths: Iterable[str]) -> List[str]:
    used: List[str] = []
    for path in sorted(paths):
        abspath = os.path.abspath(path)
        if os.path.exists(abspath):
            if any(is_parent(abspath, usedpath) for usedpath in used):
                continue
            used.append(abspath)

    return used


def get_paths(
    ro_paths: AbstractSet[str], rw_paths: AbstractSet[str]
) -> Tuple[Set[str], Set[str], Set[str]]:
    result_ro: Set[str] = set()
    result_rw: Set[str] = set()
    symlinks = set()

    def resolve_chain(path: str, path_set: Set[str]):
        chain = resolve_paths(path)
        if chain:
            # Elements are symlinks
            # the last is the end of the symlink chain.
            if os.path.islink(path):
                symlinks.add(path)

            path_set.add(os.path.realpath(path))
            for link in chain[:-1]:
                symlinks.add(link)
            # Ignore last element in chain, since it is
            # part of the resolved path (but not necessarily identical to the resolved
            # path if it's a directory symlink, but the original path pointed to either
            # a subdirectory of the symlink, or a file)
        else:
            path_set.add(path)

    for path in _dedup_subpaths(rw_paths):
        resolve_chain(path, result_rw)

    for path in _dedup_subpaths(ro_paths | rw_paths):
        if path not in result_rw:
            resolve_chain(path, result_ro)

    filtered_symlinks: Set[str] = set()
    for path in _dedup_subpaths(result_ro | result_rw | symlinks):
        if path in symlinks:
            filtered_symlinks.add(path)
    return result_ro, result_rw, filtered_symlinks


def sandbox_execute(
    command: List[str],
    permissions: Permissions,
    *,
    workdir: str,
    pipe_output: bool = False,
    profile: bool = True,
    **kwargs,
) -> subprocess.Popen:
    """
    Executes command within a sandbox

    On Windows/Sandboxie, this will start in a temporary directory rather than workdir,
    and must change directory itself, due to needing to start the process using ForceFolder,
    which can affect other processes run in the directory.

    args:
        command: The command to execute
        permissions: The permissions to set up the sandbox
        pipe_output: Whether or not output should go to a pipe in the returned process
                     object instead of stdout/stderr
        workdir: The starting directory of the process
        profile: If true, sets profile variables in the sandbox
        kwargs: Additional keyword arguments to pass to subprocess.Popen

    returns:
        Process object for the started process.
        This command does not block and wait for the process.
    """

    def cleanup():
        pass

    rw_paths = set(permissions.rw_paths)
    ro_paths = set(permissions.ro_paths)
    if env.TESTING:
        # Add source directory to whitelist so that we can get coverage from subprocesses
        rw_paths.add(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

    environ = os.environ.copy()

    if profile:
        # Set profile variables as environment variables in the sandbox
        for key, value in get_config().items():
            if isinstance(value, str):
                environ[key] = value
            elif isinstance(value, bool):
                if value:
                    environ[key] = str(value)

    if permissions.tmp:
        environ["TMP"] = permissions.tmp
        os.makedirs(permissions.tmp, exist_ok=True)

    # ensure paths added at runtime are available
    if "PYTHONPATH" not in environ:
        environ["PYTHONPATH"] = ""
    if sys.platform == "win32":
        environ["PYTHONPATH"] += ";" + ";".join(sys.path)
    else:
        environ["PYTHONPATH"] += ":" + ":".join(sys.path)

    if sys.platform == "linux":
        if permissions.global_read:
            ro_paths.add("/")
        else:
            ro_paths |= {
                "/bin",
                "/etc",
                "/gnu",
                "/lib",
                "/lib32",
                "/lib64",
                "/run",
                "/nix",
                "/opt",
                "/usr",
                "/var",
            }

        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        ro_bind: List[str] = sum([["--ro-bind", path, path] for path in ro_paths], [])
        bind: List[str] = sum([["--bind", path, path] for path in rw_paths], [])
        symlink: List[str] = sum(
            [["--symlink", os.readlink(path), path] for path in symlinks], []
        )

        # Needs to be a list of command arguments
        # I.e. NOT a string
        if profile:
            user = get_config_value("SANDBOX_PERMISSIONS", [])
            if not isinstance(user, list):
                raise TypeError(
                    "SANDBOX_PERMISSIONS must be a list of arguments to bwrap"
                )
        else:
            user = []

        sandbox_command = (
            [_required_exe("bwrap")]
            + ro_bind
            + bind
            + symlink
            + [
                "--new-session",
                "--dev",
                "/dev",
                "--proc",
                "/proc",
                "--unshare-all",
                "--chdir",
                workdir,
                "--die-with-parent",
            ]
            + user
        )
        if permissions.network:
            sandbox_command.append("--share-net")
        if permissions.tmp:
            sandbox_command.extend(["--bind", permissions.tmp, permissions.tmp])
        else:
            os.makedirs(env.PYBUILD_TMP_DIR, exist_ok=True)
            sandbox_command.extend(["--bind", env.PYBUILD_TMP_DIR, env.PYBUILD_TMP_DIR])
            environ["TMP"] = env.PYBUILD_TMP_DIR

    elif sys.platform == "win32":
        SINI = _required_exe("sbieini.exe")
        start_exe = _required_exe("start.exe")

        SBIE_LOCK.acquire()
        BOXNAME = "Portmod"

        # We use a tmeporary directory which is otherwise unused to launch the sandboxed process
        # The process must then chdir to the actual working directory.
        launchdir = TemporaryDirectory()
        workdir = launchdir.name

        def win32_cleanup():
            try:
                subprocess.check_call(  # nosec B603
                    f'"{start_exe}" /box:{BOXNAME} /silent /nosbiectrl delete_sandbox'
                )
            except subprocess.CalledProcessError:
                pass
            finally:
                # Remove Sandboxie.ini section for box
                subprocess.check_call(f'"{SINI}" set {BOXNAME} * ""')  # nosec B603
                SBIE_LOCK.release()
                launchdir.cleanup()

        cleanup = win32_cleanup  # noqa

        def add_command(command: str, typ: str, value: str, boxname=BOXNAME):
            subprocess.check_call(  # nosec B603
                f'"{SINI}" {command} {boxname} {typ} "{value}"'
            )

        # This enables a kernel-mode Windows Filtering Platform mechanism for
        # controlling network access, which should be more secure than the default
        # It requires reloading the driver, but we'll just enable it and hope the user
        # reloads at some point.
        add_command("set", "NetworkEnableWFP", "y", boxname="GlobalSettings")
        if permissions.network:
            add_command("set", "AllowNetworkAccess", "y")
        else:
            add_command("set", "NetworkAccess", "*,Block")
            add_command("set", "AllowNetworkAccess", "n")
            add_command("set", "NotifyInternetAccessDenied", "n")

        # Create a temporary directory that can be used without affecting the user's system
        if not permissions.tmp:
            environ["TMP"] = env.PYBUILD_TMP_DIR
        add_command("append", "WriteFilePath", environ["TMP"])

        # Required for use in git bash
        add_command("append", "OpenPipePath", r"\Device\NamedPipe\msys")

        ro_paths.add(os.path.normpath(os.environ["programfiles(x86)"]))
        ro_paths.add(os.path.normpath(os.environ["programfiles"]))

        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        if not permissions.global_read:
            add_command("append", "ClosedFilePath", get_personal())

        for path in rw_paths:
            add_command("append", "OpenPipePath", path)

        # A wrapper command is used, combined with ForceProcess, as starting the process
        # using sandboxie's start.exe would not give us any stdout or stderr.
        add_command("set", "ForceFolder", workdir)

        # Needs to be a list of command argument tuples
        if profile:
            user = get_config_value("SANDBOX_PERMISSIONS", [])
            if not isinstance(user, list):
                raise TypeError(
                    "SANDBOX_PERMISSIONS must be a list with sbieini commands in the order of $command $type $value. "
                    'E.g. ["append", "ClosedFilePath", "filename.txt"]'
                )
            for user_command, typ, value in user:
                add_command(user_command, typ, value)
        add_command("set", "Enabled", "y")

        sandbox_command = []
    elif sys.platform == "darwin":
        sandbox_command = [_required_exe("sandbox-exec"), "-p"]

        sandbox_profile = """
            (version 1)
            (deny default)
            (allow process-exec*)
            (allow signal (target self))
            (allow process-fork)
            (allow sysctl-read)
            (allow file-read*
                file-write-data
                (literal "/dev/null")
                (literal "/dev/zero")
            )
            (allow file-read*
                file-write-data
                file-ioctl
                (literal "/dev/dtracehelper")
            )
            (allow ipc-posix-shm)
        """

        if permissions.global_read:
            sandbox_profile += """
                (allow file-read*)
            """
        else:
            sandbox_profile += """
                (allow file-read-data file-read-metadata
                  (regex "^/dev/autofs.*")
                  (regex "^/Users/[^/]*/.gitconfig")
                  (literal "/var")
                  (literal "/dev/fd")
                  (literal "/dev/random")
                  (literal "/dev/urandom")
                  (literal "/var/db")
                  (literal "/var/db/xcode_select_link")
                  (literal "/")
                )
            """

            ro_paths |= {
                "/usr",
                "/System/Library",
                "/Applications",
                "/etc",
                "/private/etc",
                "/Library",
                "/opt",
            }

        if not permissions.tmp:
            environ["TMPDIR"] = env.PYBUILD_TMP_DIR
            os.makedirs(environ["TMPDIR"], exist_ok=True)
        rw_paths.add(environ["TMPDIR"])
        ro_paths.add("/private/var/select/sh")
        ro_paths, rw_paths, symlinks = get_paths(ro_paths, rw_paths)

        if not permissions.global_read:
            sandbox_profile += """
                (allow file-read*"""
            for path in ro_paths:
                sandbox_profile += f"""
                    (subpath "{path}")"""
            for path in parents(ro_paths | rw_paths):
                sandbox_profile += f"""
                    (literal "{path}")"""
            for path in symlinks:
                sandbox_profile += f"""
                    (literal "{path}")"""
            sandbox_profile += """
                )"""

        if permissions.network:
            sandbox_profile += "\n(allow network*)"

        sandbox_profile += """
            (allow file-write* file-read*"""
        for path in rw_paths:
            sandbox_profile += f"""
                (subpath "{path}")"""
        sandbox_profile += """
            )"""

        # Must be a string
        if profile:
            user = get_config_value("SANDBOX_PERMISSIONS", "")
            if not isinstance(user, str):
                raise TypeError(
                    "SANDBOX_PERMISSIONS must be a string with sandbox-exec profile contents"
                )
            sandbox_profile += user

        sandbox_command.append(sandbox_profile)
    else:
        raise Exception("Unsupported Platform")

    try:
        kwargs = dict(kwargs)
        if pipe_output:
            # Pipe stdout, and redirect stderr to the same pipe
            kwargs["stdout"] = subprocess.PIPE
            kwargs["stderr"] = subprocess.STDOUT

        proc = subprocess.Popen(  # nosec B603
            sandbox_command + command,
            env=environ,
            encoding="utf-8",
            cwd=workdir,
            **kwargs,
        )

        def wait_and_cleanup():
            proc.wait()
            cleanup()

        Thread(target=wait_and_cleanup).start()
    except Exception as err:
        cleanup()
        raise err

    return proc



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, Iterable, List, Optional, Tuple

from portmod.config.mask import get_unmasked
from portmod.globals import env
from portmod.parsers.flags import add_flag, get_flags, remove_flag
from portmodlib.atom import QualifiedAtom, atom_sat
from portmodlib.version import Version, VersionRange

from ..config import get_config
from ..pybuild import Pybuild


# Note: in KEYWORDS, the absence of a keyword indicates untested
# The UNTESTED variant is never associated with a NamedKeyword
class Stability(Enum):
    MASKED = "masked"
    TESTING = "testing"
    STABLE = "stable"
    UNTESTED = "untested"

    def __str__(self):
        return self.value

    def __le__(self, other: Any):
        return self < other or self == other

    def __lt__(self, other: Any):
        if not isinstance(other, Stability):
            raise NotImplementedError()

        order = [
            Stability.MASKED,
            Stability.UNTESTED,
            Stability.TESTING,
            Stability.STABLE,
        ]
        return order.index(self) < order.index(other)


# TODO: Phase out '.'
_kw_name_re = re.compile(r"^([A-Za-z0-9][A-Za-z0-9_.-]*)")


class Keyword:
    """Keyword indicating the stability/visibility of a package on various architectures"""

    def accepts(self, _other: "Keyword") -> bool:
        """
        Returns true if and only if this keyword accepts the other keyword
        This is not necessarily true in reverse (~arch accepts arch, but not vice-versa)
        """
        raise NotImplementedError("This must be implemented for keyword subclasses")

    def masks(self, _arch: str, _arch_ver: Optional[Version] = None) -> bool:
        """
        Returns true if and only if this keyword masks the other keyword
        """
        raise NotImplementedError("This must be implemented for keyword subclasses")


def parse_keyword(inputstr: str) -> Keyword:
    """
    Converts the string representation of a Keyword

    The Keyword class itself is just an abstract class. This function, or the
    WildcardKeyword and NamedKeyword classes, shuold be used to instantiate keywords.
    """
    try:
        return WildcardKeyword(inputstr)
    except ValueError:
        return NamedKeyword.from_string(inputstr)


class WildcardKeyword(Keyword, Enum):
    """Wildcards used to control package visiblily on multiple architectures"""

    ALWAYS = "**"
    """Package is always visible"""
    TESTING = "~*"
    """Package is visible if testing on any architecture"""
    STABLE = "*"
    """Package is visible if stable on any architecture"""
    MASKED = "-*"
    """Package is masked on all architectures not otherwise listed"""

    def __str__(self):
        return self.value

    def masks(self, _arch: str, _arch_ver: Optional[Version] = None) -> bool:
        if self == WildcardKeyword.MASKED:
            return True
        return False

    def accepts(self, other: Keyword) -> bool:
        if WildcardKeyword.MASKED in (self, other):
            return False
        if WildcardKeyword.ALWAYS in (self, other):
            return True
        if isinstance(other, WildcardKeyword):
            if self == WildcardKeyword.STABLE:
                return other == WildcardKeyword.STABLE
            if self == WildcardKeyword.TESTING:
                return other == WildcardKeyword.TESTING
        elif isinstance(other, NamedKeyword):
            if self == WildcardKeyword.STABLE:
                return other.stability in (Stability.STABLE, Stability.TESTING)
            if self == WildcardKeyword.TESTING:
                return other.stability == Stability.TESTING

        raise NotImplementedError()


@dataclass
class NamedKeyword(Keyword):
    """
    Keyword referring to a specific architecture

    (and optionally specific versions of that architecture)
    """

    name: str
    stability: Stability
    versions: Optional[VersionRange]

    @classmethod
    def from_string(cls, inputstr: str) -> "NamedKeyword":
        """
        Constructs a keyword from the string representation

        The string representation should consist of the keyword name,
        optionally preceeded by a ~ or - to indicate testing/masked keywords
        and optionally followed by a comma-separated list of versions enclosed in braces

        E.g. ~openmw{==0.48}
        """
        if inputstr[0] == "~":
            stability = Stability.TESTING
            inputstr = inputstr[1:]
        elif inputstr[0] == "-":
            stability = Stability.MASKED
            inputstr = inputstr[1:]
        else:
            stability = Stability.STABLE

        result = re.split(_kw_name_re, inputstr, maxsplit=1)
        if len(result) < 3:
            raise ValueError(f"Invalid Keyword '{inputstr}'!")

        _, name, versions = result

        if versions:
            if (
                not versions.startswith("{") or not versions.endswith("}")
            ) and versions[0].isnumeric():
                raise ValueError(
                    f"Missing braces around keyword versions in keyword {inputstr}"
                )
            if not versions.startswith("{"):
                raise ValueError(
                    f"Trailing characters {versions} of keyword {inputstr} "
                    "cannot be part of a valid keyword name"
                )
            versions = VersionRange(versions.strip("{}"), external=True)
        else:
            versions = None
        return cls(name, stability, versions)

    def __str__(self):
        result = self.name
        if self.stability == Stability.TESTING:
            result = "~" + result
        elif self.stability == Stability.MASKED:
            result = "-" + result
        if self.versions:
            result += "{" + str(self.versions) + "}"
        return result

    def masks(self, arch: str, arch_ver: Optional[Version] = None) -> bool:
        if self.stability != Stability.MASKED:
            return False

        if arch != self.name:
            return False

        # If the mask does not specify a version, it always applies
        if self.versions is None:
            return True
        # If the mask specifies a version, but no version is provided, assume it doesn't apply
        # This is a situation which should usually not occur,
        # as packages should only provide a version if versioning is set up
        elif arch_ver is None:
            return False

        # Otherwise, the mask applies only if the version is within the masked range
        return arch_ver in self.versions

    def accepts(self, other: Keyword) -> bool:
        if isinstance(other, NamedKeyword):
            if other.name != self.name:
                return False

            # If one is masked, do not accept
            if (
                other.stability == Stability.MASKED
                and self.stability != Stability.MASKED
            ):
                return False
            if (
                self.stability == Stability.MASKED
                and other.stability != Stability.MASKED
            ):
                return False

            # Stable does not accept testing, but testing accepts stable
            if (
                other.stability == Stability.TESTING
                and self.stability == Stability.STABLE
            ):
                return False

            # If one or the other does not specify a version, there is overlap
            if other.versions is None or self.versions is None:
                return True

            # Otherwise, there is only overlap if the versions overlap
            return bool(self.versions & other.versions)

        elif isinstance(other, WildcardKeyword):
            return other.accepts(self)

        raise NotImplementedError()


def _user_package_accept_keywords_path() -> str:
    return os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")


def add_keyword(
    atom: QualifiedAtom, keyword: Keyword, *, protect_file: Optional[str] = None
):
    """Adds keyword for the given atom. Does not modify any existing keywords."""
    keyword_file = protect_file or _user_package_accept_keywords_path()

    add_flag(keyword_file, atom, str(keyword))


def remove_keyword(atom: QualifiedAtom, keyword: Keyword):
    keyword_file = os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    remove_flag(keyword_file, atom, str(keyword))


def get_global_keywords() -> List[Keyword]:
    return [parse_keyword(token) for token in get_config()["ACCEPT_KEYWORDS"]]


def get_keywords(atom: QualifiedAtom) -> List[Keyword]:
    """Returns the keywords for a particular package"""
    keyword_file = os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    return [
        parse_keyword(token) for token in get_flags(keyword_file, atom)
    ] + get_global_keywords()


def _accepts(accept_keywords: Iterable[Keyword], keywords: Iterable[Keyword]):
    # FIXME: This description is rediculous.
    """
    Returns true if and only if at least one keyword in accept_keywords
    accepts at least one of the keywords in keywords

    You must not pass a consumable iterator to this method, as it will iterate
    over them multiple times.
    """
    for accept_keyword in accept_keywords:
        if isinstance(accept_keyword, WildcardKeyword):
            if accept_keyword == WildcardKeyword.STABLE:
                # Accepts stable on all architectures. Valid if keywords contains
                # a stable keyword for any keyword
                if any(
                    keyword.stability == Stability.STABLE
                    for keyword in keywords
                    if isinstance(keyword, NamedKeyword)
                ):
                    return True
            elif accept_keyword == WildcardKeyword.TESTING:
                # Accepts testing on all architectures. Valid if keywords contains
                # either testing or stable for any keyword
                if any(
                    isinstance(keyword, NamedKeyword)
                    and keyword.stability == Stability.TESTING
                    for keyword in keywords
                ):
                    return True
            elif accept_keyword == WildcardKeyword.ALWAYS:
                # Accepts any configuration
                return True
        elif isinstance(accept_keyword, NamedKeyword):
            if any(accept_keyword.accepts(keyword) for keyword in keywords):
                return True

    return WildcardKeyword.ALWAYS in keywords


def get_stability(package: Pybuild) -> Tuple[Stability, Optional[Keyword]]:
    """Returns the stability for the user's current configuration of the given package"""
    accepts_keywords = get_keywords(package.ATOM)
    keywords = list(map(parse_keyword, package.KEYWORDS))

    is_unmasked = False
    for atom in get_unmasked().get(package.CPN, []):
        if atom_sat(atom, package.ATOM):
            is_unmasked = True
    arch_ver = get_config().get("ARCH_VERSION")
    return _get_stability(
        accepts_keywords,
        keywords,
        get_config()["ARCH"],
        Version(arch_ver) if arch_ver else None,
        is_unmasked,
    )


def _get_stability(
    accepts_keywords: List[Keyword],
    keywords: List[Keyword],
    arch: str,
    arch_ver: Optional[Version] = None,
    is_unmasked: bool = False,
) -> Tuple[Stability, Optional[Keyword]]:
    # Maked named keywords take priority. To override, you need to use package.unmask
    if not is_unmasked:
        for keyword in keywords:
            if isinstance(keyword, NamedKeyword) and keyword.masks(arch, arch_ver):
                return Stability.MASKED, keyword

    # Note: if a testing keyword is in accepts_keywords,
    # we treat packages with matching testing keywords as stable
    if _accepts(accepts_keywords, keywords):
        return Stability.STABLE, None

    for keyword in accepts_keywords:
        # Stable keyword implies testing keyword
        if isinstance(keyword, NamedKeyword):
            if keyword.stability == Stability.STABLE:
                testing_keyword = NamedKeyword(
                    keyword.name, Stability.TESTING, keyword.versions
                )
                for keyword in keywords:
                    if testing_keyword.accepts(keyword):
                        return Stability.TESTING, keyword

    if not is_unmasked:
        for keyword in keywords:
            if isinstance(keyword, WildcardKeyword) and keyword.masks(arch, arch_ver):
                return Stability.MASKED, keyword

    # Untested
    return Stability.UNTESTED, None



from test.env import setup_env
from test.merge import merge

import pytest
from pytest import fail

from portmod.loader import load_installed_pkg
from portmodlib.atom import Atom

try:
    import PySide6  # noqa: F401

    from portmod._gui.packages import get_installed_packages, get_local_flags
except ModuleNotFoundError as error:
    if error.name == "PySide6":
        pytest.skip("GUI tests when PySide6 is not installed", allow_module_level=True)
    raise error


def test_get_installed_packages():
    setup_env("test")
    merge(["test/test0"], nodeps=True)
    merge(["test/test2"], nodeps=True)
    merge(["test/test3"], nodeps=True)

    packages = {
        load_installed_pkg(Atom("test/test0")),
        load_installed_pkg(Atom("test/test2")),
        load_installed_pkg(Atom("test/test3")),
    }

    assert set(get_installed_packages()) == packages


def test_get_local_flags():
    setup_env("test")
    merge(["test/test4"], nodeps=True)

    installed_pybuild = load_installed_pkg(Atom("test/test4"))
    if installed_pybuild:
        assert get_local_flags(installed_pybuild) == {"baf": (False, "test flag")}
    else:
        fail()



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Abstract interface for IO

This currently only targets the merge function in the portmod/merge.py file via the
MergeIO abstract class, and install_pkg via InstallIO (install_pkg gets called by
merge, but shouldn't be invoked by itself).

To use, call merge(..., io=my_merge_io), passing an object from a class which
implements MergeIO as my_merge_io.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    AbstractSet,
    Callable,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Set,
    TypeVar,
    cast,
)

from portmod.loader import SandboxedProcess
from portmod.pybuild import Pybuild
from portmod.util import KeywordDep, LicenseDep
from portmodlib.atom import Atom, FQAtom

from .util import UseDep


class Transaction:
    """Transaction class"""

    REPR: str
    COLOUR: Callable[[str], str]
    pkg: Pybuild
    flags: Set[str]

    def __init__(self, pkg: Pybuild, flags: Iterable[str]):
        self.pkg = pkg
        self.flags = set(flags)

    def __str__(self):
        return f"{self.__class__.__name__}({self.pkg})"

    def __repr__(self):
        return str(self)


_RES = TypeVar("_RES")


class Task(Generic[_RES]):
    """A function which the interface is expected to call"""

    def __init__(self, func: Callable[..., _RES], *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.done: bool = False
        self._result: Optional[_RES] = None

    def result(self) -> _RES:
        if self.done:
            return cast(_RES, self._result)
        raise RuntimeError("Cannot get the result of a task which has not run!")

    def run(self):
        """Runs the task"""
        if not self.done:
            self._result = self.func(*self.args, **self.kwargs)
            self.done = True
        else:
            raise Exception("Tasks cannot be run twice!")


@dataclass(frozen=True)
class Message:
    """A message produced by a package"""

    class Type(Enum):
        """The type of the package's message"""

        INFO = "info"
        WARNING = "warning"

    typ: Type
    text: str


class MergeMode(Enum):
    """
    Merge modes. Value corresponds to the localised text
    """

    INSTALL = "to-install"
    REMOVE = "to-remove"


class PhaseFunction(Enum):
    SRC_UNPACK = "src_unpack"
    SRC_PREPARE = "src_prepare"
    SRC_INSTALL = "src_install"
    PKG_POSTINST = "pkg_postinst"
    PKG_PRERM = "pkg_prerm"


class Progress(ABC):
    """
    Abstract class for receiving information which can be displayed in progressbars
    """

    def __init__(self, steps: int):
        self.steps = steps

    @abstractmethod
    def update(self, *, value: Optional[int] = None): ...

    @abstractmethod
    def done(self): ...


class PackageIO(ABC):
    """
    Abstract class for implementing input/output handling for :py:func:`portmod.package:install_pkg`
    and :py:func:`portmod.package.remove_pkg`
    """

    def __init__(self, package: FQAtom, pipe_output: bool):
        """
        args:
            package: The package atom, fully qualified
            pipe_output: Whether or not stdout and stderr should be piped. If true, no
                         stderr or stdout will be produced, and the implementor will be
                         required to read and handle this using Popen.communicate
        """
        self.package = package
        self.pipe_output = pipe_output

    @abstractmethod
    def phase_function(self, function: PhaseFunction, process: SandboxedProcess):
        """
        A phase function has begun execution

        The process can be monitored to determine when execution has ended,
        and its communicate function can be used to read output (if pipe_output was set)

        Note that the subprocess will run in the background, and the install process will
        wait on it after this function has finished executing.
        The purpose of this function is to allow a UI to monitor and use the phase function's
        output in real time.
        """


class RemoveIO(PackageIO):
    @abstractmethod
    def remove_files(self, count: int) -> Progress:
        """Should return something which can handle displaying file installation progress"""

    @abstractmethod
    def begin_removal(self):
        """Package removal has begun"""

    @abstractmethod
    def finished_removal(self):
        """Package removal has finished successfully"""


class InstallIO(RemoveIO):
    """
    Interface for installing packages.

    The old package, if present, will be removed as part of installation via the RemoveIO interface
    """

    @abstractmethod
    def check_conflicts(self, count: int) -> Progress:
        """Should return something which can handle displaying file conflict progress"""

    @abstractmethod
    def install_files(self, count: int) -> Progress:
        """Should return something which can handle displaying file installation progress"""

    @abstractmethod
    def begin_install(self):
        """Package installation has begun"""

    @abstractmethod
    def finished_install(self):
        """Package installation has finished successfully"""

    @abstractmethod
    def can_overwrite(self, path: str) -> bool:
        """
        Should return true if the specified path can be overwritten

        Is only called on files which already exist on the filesystem and are not owned by another package
        """


class MergeIO(ABC):
    """
    Abstract class for implementing input/output handling for :py:func:`portmod.merge:configure`
    """

    @abstractmethod
    def display_transactions(
        self,
        mode: MergeMode,
        transactions: Sequence[Transaction],
        new_selected: AbstractSet[Pybuild],
    ):
        """
        Transaction list to display to the user

        args:
            transactions: List of transactions to perform, in order.
            new_selected: Packages which were selected for installation
                          these, along with previously selected packages
                          (use :py:func:`portmod.config.sets:is_selected`) should be
                          visually distinguished from packages which are installed
                          just as dependencies
        """

    @abstractmethod
    def use_changes(self, use_flags: Sequence[UseDep], apply_callback: Task):
        """
        Display necessary flag changes to the user and prompt them to accept
        the changes

        args:
            use_flags: he use flag changes which are required
            apply_callback: Task, which when executed will apply the changes
        """

    @abstractmethod
    def keyword_changes(self, keywords: Sequence[KeywordDep], apply_callback: Task):
        """
        Display necessary keyword changes to the user and prompt them to accept
        the changes

        args:
            keywords: The keyword changes which are required
            apply_callback: Task, which when executed will apply the changes
        """

    @abstractmethod
    def license_changes(self, licenses: Sequence[LicenseDep], apply_callback: Task):
        """
        Display licenses which need to be accepted and prompt the user to accept them

        args:
            licenses: The licenses which must be accepted
            apply_callback: Task, which when executed will apply the changes
        """

    @abstractmethod
    def pkg_nofetch(self, package: FQAtom, instructions: str):
        """
        Display fetch information for the package

        args:
            package: The package which contains files which could not be fetched
            instructions: The instructions produced by the package describing how
                         to manually fetch the files
        """

    @abstractmethod
    def pkg_pretend(self, package: FQAtom, message: str):
        """
        Display pkg_pretend messages for the package

        args:
            package: The package which produced the message
            message: The output from pkg_pretend
        """

    @abstractmethod
    def space_warning(self, package: FQAtom, message: str):
        """
        Display message about insufficient space

        args:
            package: The package which contains files which could not be fetched
            message: Message to display
        """

    @abstractmethod
    def rebuild_warning(self, packages: Sequence[Atom], message: str):
        """
        Display message about packages which need to be rebuilt

        args:
            packages: The package which need to be rebuilt
            message: Message to display
        """

    @abstractmethod
    def download_ready(self, start_callback: Task):
        """
        Called when all checks have passed and we are ready to download files

        args:
            start_callback: Task to indicate that the user wants to begin merge
        """

    @abstractmethod
    def merge_ready(self, start_callback: Task):
        """
        Called when files have been downloaded and we are ready to merge
        User is expected to handle manual downloads before this point
        May be called again if manually downloaded files do not pass checks

        args:
            start_callback: Task to indicate that the user wants to begin merge
        """

    @abstractmethod
    def pkg_messages(self, package: FQAtom, messages: Sequence[Message]):
        """
        Display custom messages for the package

        atgs:
            package: The package which produced the custom messages
            messages: Messages to display
        """

    @abstractmethod
    def get_install_io(self, package: FQAtom) -> InstallIO:
        """
        Produce an InstallIO object for use when installing a package

        This will be called once for each package to be installed
        """

    @abstractmethod
    def get_remove_io(self, package: FQAtom) -> RemoveIO:
        """
        Produce an RemoveIO object for use when removing a package

        This will be called once for each package to be removed
        """

    @abstractmethod
    def finished(self, message: str):
        """
        Called when the merge operation has finished. The message should be displayed to the user
        """

    @abstractmethod
    def get_stepped_fetch_progress(
        self, step: int, max_steps: int
    ) -> Callable[[str, int, Optional[int]], Progress]:
        """
        Produces a method that generates progress bars for displaying file fetching
        annotated with the step number, and maximum number of steps
        """



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import re
from typing import AbstractSet, Any, Dict, Optional, Set, TypeVar

from .version import Version, ver_re

"""
Module for handling package atoms

All atom classes defined in this module should be considered read-only.
Modification of an Atom object may have unexpected side-effects
"""

flag_re = r"[A-Za-z0-9][A-Za-z0-9+_-]*"
useflag_re = re.compile(r"^" + flag_re + r"$")
usedep_re = (
    r"(?P<prefix>[!-]?)(?P<flag>"
    + flag_re
    + r")(?P<default>(\(\+\)|\(\-\))?)(?P<suffix>[?=]?)"
)
_usedep_re = re.compile("^" + usedep_re + "$")

op_re = r"(?P<B>(!!))?(?P<OP>([<>]=?|[<>=~]))?"
cat_re = r"((?P<C>[A-Za-z0-9][A-Za-z0-9\-]*)/)?"
repo_re = r"(::(?P<R>[A-Za-z0-9_][A-Za-z0-9_-]*(::installed)?))?"
_atom_re = re.compile(
    op_re
    + cat_re
    + r"(?P<P>(?P<PN>[A-Za-z0-9+_-]+?)(-(?P<PV>"
    + ver_re
    + r"))?)"
    + repo_re
    + r"(\[(?P<USE>.*)\])?$"
)


class InvalidAtom(Exception):
    "Exception indicating an atom has invalid syntax"


class UnqualifiedAtom(Exception):
    """
    Exception indicating an atom, which was expected to be
    qualified with a category, has no category
    """

    def __init__(self, atom):
        self.atom = atom

    def __str__(self):
        return f"Atom {self.atom} was expected to have a category!"


T = TypeVar("T", bound="Atom")


class Atom(str):
    @property
    def CP(self) -> Optional[str]:
        """
        The category, package name and version, excluding revision

        E.g. ``base/example-suite-1.0``
        """
        if self.CPN is None:
            return None
        if self.version is not None:
            return self.CPN + "-" + self.version.display(revision=True)
        return self.CPN

    CPN: Optional[str]
    """
    The category and package name

    E.g. ``base/example-suite``
    """
    USE: Set[str] = set()
    """
    Use flags set on the atom.

    E.g. ``{minimal}`` in ``base/example-suite[minimal]``
    """

    @property
    def P(self) -> str:
        """
        The package name and version, excluding revision

        E.g. ``example-suite-1.0``
        """
        return (
            self.PN
            + "-"
            + (self.version.display(revision=False) if self.version is not None else "")
        )

    PN: str
    """
    The package name

    E.g. ``example-suite``
    """

    @property
    def PF(self) -> str:
        """
        The package name and version, including revision

        E.g. ``example-suite-1.0-r1``
        """
        return self.PN + (
            "-" + self.version.display() if self.version is not None else ""
        )

    @property
    def PV(self) -> Optional[str]:
        """
        The package version, not including revision

        E.g. ``1.0``
        """
        if self.version is not None:
            return self.version.display(revision=False)
        return None

    @property
    def PR(self) -> Optional[str]:
        """
        The package revision

        E.g. ``r1``
        """
        if self.version is not None:
            return f"r{self.version.revision}"
        return None

    C: Optional[str]
    """
    The package category

    E.g. ``base``
    """
    R: Optional[str]
    """
    The package repository

    E.g. ``openmw`` in base/example-suite-1.0-r1::openmw
    """
    OP: Optional[str]
    """
    An operator.

    See PMS section `8.2.6.1 <https://projects.gentoo.org/pms/7/pms.html#x1-760008.2.6.1>`_,
    noting that Weak blocks are not currently supported by Portmod, and blocks don't show up
    in OP (they instead set the BLOCK field).

    E.g. ``openmw`` in base/example-suite-1.0-r1::openmw
    """
    BLOCK: bool
    """
    If true, this atom is a blocker, referring to a package which must not be installed.

    E.g. ``!!base/example-suite``
    """

    @property
    def PVR(self) -> Optional[str]:
        """
        The version and revision

        E.g. 1.0-r1
        """
        if self.version is not None:
            return self.version.display(revision=True)
        return None

    @property
    def CPF(self) -> str:
        """
        The category, package name and version, including revision

        E.g. ``base/example-suite-1.0-r1``
        """
        if self.CPN is None:
            prefix = self.PN
        else:
            prefix = self.CPN
        if self.version is not None:
            return prefix + "-" + self.version.display(revision=True)
        return prefix

    version: Optional["Version"] = None
    """
    The package's version
    """

    _CACHE: Dict[str, Any] = {}

    def __init__(self, atom: str):
        if atom in self._CACHE:
            self.__dict__ = self._CACHE[atom]
            return

        match = _atom_re.match(atom)
        if not match:
            raise InvalidAtom("Invalid atom %s. Cannot parse" % (atom))

        if match.group("PN") and match.group("C"):
            self.CPN = match.group("C") + "/" + match.group("PN")
        else:
            self.CPN = None

        if match.group("USE"):
            self.USE = set(match.group("USE").split(","))
            for x in self.USE:
                m = _usedep_re.match(x)
                if not m:
                    raise InvalidAtom(
                        "Invalid use dependency {} in atom {}".format(atom, x)
                    )

        if match.group("PV"):
            self.version = Version("", _match=match)

        self.PN = match.group("PN")
        self.C = match.group("C")
        self.R = match.group("R")
        self.OP = match.group("OP")
        self.BLOCK = match.group("B") is not None

        if self.OP is not None and self.version is None:
            raise InvalidAtom(
                "Atom %s has a comparison operator but no version!" % (atom)
            )

        self._CACHE[atom] = self.__dict__

    def evaluate_conditionals(self, use: AbstractSet[str]) -> "Atom":
        """
        Create an atom instance with any USE conditionals evaluated.
        @param use: The set of enabled USE flags
        @return: an atom instance with any USE conditionals evaluated
        """
        tokens = set()

        for x in self.USE:
            m = _usedep_re.match(x)

            if m is not None:
                operator = m.group("prefix") + m.group("suffix")
                flag = m.group("flag")
                default = m.group("default")
                if default is None:
                    default = ""

                if operator == "?":
                    if flag in use:
                        tokens.add(flag + default)
                elif operator == "=":
                    if flag in use:
                        tokens.add(flag + default)
                    else:
                        tokens.add("-" + flag + default)
                elif operator == "!=":
                    if flag in use:
                        tokens.add("-" + flag + default)
                    else:
                        tokens.add(flag + default)
                elif operator == "!?":
                    if flag not in use:
                        tokens.add("-" + flag + default)
                else:
                    tokens.add(x)
            else:
                raise Exception("Internal Error when processing atom conditionals")

        atom = Atom(self)
        atom.USE = tokens
        return atom

    def strip_use(self: T) -> T:
        """Returns the equivalent of this atom with the USE dependencies removed"""
        return self.__class__(re.sub(r"\[.*\]", "", str(self)))

    def use(self, *flags: str):
        """returns atom with use flag dependency"""
        return self.__class__(f'{self}[{",".join(flags)}]')

    def with_category(self, category: str) -> "QualifiedAtom":
        return QualifiedAtom(
            (self.BLOCK and "!!" or "")
            + (self.OP or "")
            + category
            + "/"
            + self.PF
            + (self.R or "")
        )


class QualifiedAtom(Atom):
    """Atoms that include categories"""

    CP: str
    CPN: str
    CPF: str
    C: str

    def __init__(self, atom: str):
        super().__init__(atom)

        if not self.C:
            raise UnqualifiedAtom(atom)


class VAtom(QualifiedAtom):
    """Atoms that include version information"""

    PV: str
    PVR: str
    version: "Version"

    def __init__(self, atom: str):
        super().__init__(atom)
        if not self.version:
            raise InvalidAtom(f"Atom {atom} should include a version")


class FQAtom(VAtom):
    """Atoms that include all possible non-optional fields"""

    R: str

    def __init__(self, atom: str):
        super().__init__(atom)
        if not self.R:
            raise InvalidAtom(f"Atom {atom} should include a repository")


def atom_sat(specific: Atom, generic: Atom, *, ignore_name: bool = False) -> bool:
    """
    Determines if a fully qualified atom (can only refer to a single package)
    satisfies a generic atom
    """

    if not ignore_name:
        if specific.PN != generic.PN:
            # Mods must have the same name
            return False

        if generic.C and (generic.C != specific.C):
            # If para defines category, it must match
            return False

    if generic.R and (generic.R != specific.R):
        # If para defines repo, it must match
        return False

    if not generic.OP:
        # Simple atom, either one version or all versions will satisfy

        # Check if version is correct
        if generic.version and (
            not specific.version
            or not specific.version.compare(generic.version, ignore_revision=True)
        ):
            return False

        # Check if revision is correct
        if (
            generic.version
            and generic.version.revision
            and (
                not specific.version
                or specific.version.revision != generic.version.revision
            )
        ):
            return False
    elif generic.version and specific.version:
        equal = specific.version == generic.version
        verequal = specific.version.compare(generic.version, ignore_revision=True)
        lessthan = specific.version < generic.version
        greaterthan = specific.version > generic.version

        if generic.OP == "=":
            return equal
        if generic.OP == "~":
            return verequal
        if generic.OP == "<":
            return lessthan
        if generic.OP == "<=":
            return equal or lessthan
        if generic.OP == ">":
            return greaterthan
        if generic.OP == ">=":
            return equal or greaterthan

    return True



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

# Necessary to allow _K, etc. to be defined in TYPE_CHECKING only
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

if TYPE_CHECKING:
    from typeshed import SupportsDunderGT, SupportsDunderLT

    _SupportsRichComparison = Union[SupportsDunderLT, SupportsDunderGT]
    _K = TypeVar("_K", bound=_SupportsRichComparison)


class CycleException(Exception):
    """Exception indicating a graph cycle"""

    def __init__(self, message, cycle: Union[List[str], List[Tuple[str]]]):
        if isinstance(cycle[0], tuple):
            cyclestr = " -> ".join(map(lambda x: x[0], cycle))
        else:
            cyclestr = " -> ".join(cast(List[str], cycle))
        super().__init__(message + " " + cyclestr)
        self.cycle = cycle


def __min_priority(priorities: Mapping[_K, _SupportsRichComparison], nodes: Set[_K]):
    priority = "z"
    resultset = set()
    for node in nodes:
        if priorities[node] < priority:
            priority = priorities[node]
            resultset = set([node])
        elif priorities[node] == priority:
            resultset.add(node)

    return resultset


class WeightedEdge(NamedTuple):
    node: _SupportsRichComparison
    weight: _SupportsRichComparison


def tsort(
    graph: Union[
        Mapping[_K, Set[_K]],
        Mapping[_K, Set[WeightedEdge]],
    ],
    priority: Optional[Mapping[_K, _SupportsRichComparison]] = None,
):
    """
    Topologically sorts the given graph using Kahn's algorithm.

    The smallest elements appear early to ensure consistency between runs.

    args:
        graph: A dictionary mapping each vertex to a set containing its children.
               Edges must either be all weighted, or all unweighted, but None
               as a weight is equivalent to an unweighted edge and will never
               be ignored to break a cycle.
        pirority: A dictionary mapping each verted to a priority (anything which can be compared)


    returns:
        A topologically sorted list of all nodes in the graph, with each node appearing
        before its children.
    """
    _graph = dict(graph)
    unweighted = None
    L: List[_K] = []

    def find_roots():
        """Finds all root nodes without incoming edges"""
        S = {node for node in _graph.keys() if node not in L}
        nonlocal unweighted
        for edges in _graph.values():
            if unweighted:
                S -= edges
            elif unweighted is False:
                for edge, _ in edges:
                    S.discard(edge)
            elif unweighted is None:
                for value in edges:
                    if isinstance(value, WeightedEdge):
                        unweighted = False
                        S.discard(value.node)
                    else:
                        unweighted = True
                        S.discard(value)
        return S

    def push_smallest(S: Set[_K], L: List[_K]):
        # We always take the smallest value from the set, rather than an arbitrary value
        if priority is None:
            smallest = min(S)
        else:
            smallest = min(__min_priority(priority, S))

        S.remove(smallest)
        L.append(smallest)
        return smallest

    while any(edges for edges in _graph.values()):
        S = find_roots()
        while len(S) > 0:
            smallest = push_smallest(S, L)

            s_set = _graph[smallest]
            _graph[smallest] = set()
            for node in s_set:
                if unweighted:
                    if not any(node in edges for edges in _graph.values()):
                        S.add(node)
                else:
                    node, weight = node
                    if not any(
                        node in {edge for edge, _ in edges} for edges in _graph.values()
                    ):
                        S.add(node)

        if any(edges for edges in _graph.values()):
            # Graph has at least one cycle

            def invert(g):
                new: Dict[Any, Any] = {}
                for node in g:
                    for link in g[node]:
                        if not unweighted:
                            link, _ = link
                        if link in new:
                            new[link].add(node)
                        else:
                            new[link] = {node}
                return new

            def search(g, node, previous):
                if node in previous:
                    i = previous.index(node)
                    return previous[i:] + [node]

                for other in g[node]:
                    result = search(g, other, previous + [node])
                    if result:
                        return result
                return None

            # Invert graph and search for cycle
            cycle = search(
                invert(_graph), next(node for node in _graph if _graph[node]), []
            )

            if unweighted:
                raise CycleException("There is a cycle in the graph!", cycle)

            # If the edges have weights, find the smallest weight edge
            # in the cycle and remove it from the graph
            # After removal, we look for nodes without incoming edges and continue
            # adding to the sorted list
            min_edge = None
            min_weight = None
            while len(cycle) > 1:
                first_node, next_node = cycle[0], cycle[1]
                for edge, weight in _graph[next_node]:
                    if (
                        edge == first_node
                        and weight is not None
                        and (min_weight is None or weight < min_weight)
                    ):
                        min_edge = first_node, next_node
                        min_weight = weight
                del cycle[0]
            if min_edge is None:
                raise CycleException("There is a cycle in the graph!", cycle)
            else:
                _graph[min_edge[1]].remove(WeightedEdge(min_edge[0], min_weight))
    else:
        # If there are no edges
        S = find_roots()
        while len(S) > 0:
            push_smallest(S, L)

    return L



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Dict, Optional, cast

from portmod.repo import BaseRepo, get_repo
from portmod.repos import get_local_repos, get_remote_repos
from portmod.sync import sync
from portmodlib.l10n import l10n


def sync_args(args):
    local_repos = get_local_repos()
    remote_repos = get_remote_repos()
    to_sync: Dict[str, BaseRepo] = {}

    meta = get_repo("meta")
    # Always sync meta, in case it needs to inform us
    # of newly added repositories
    if not os.path.exists(meta.location):
        sync([meta])
    else:
        to_sync["meta"] = meta

    for name in args.repository or local_repos:
        repo = cast(Optional[BaseRepo], local_repos.get(name) or remote_repos.get(name))
        if repo is None:
            raise RuntimeError(f"Unknown repository {name}")
        to_sync[name] = repo

    sync(to_sync.values())


def add_sync_parser(subparsers, parents):
    parser = subparsers.add_parser("sync", help=l10n("sync-help"), parents=parents)
    parser.add_argument(
        "repository",
        help=l10n("sync-repositories-help"),
        nargs="*",
    )
    parser.set_defaults(func=sync_args)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Various utility functions
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, List, NamedTuple, Optional, Tuple

from portmod.config.license import has_eula, is_license_accepted
from portmod.repo.keywords import (
    Keyword,
    NamedKeyword,
    Stability,
    WildcardKeyword,
    get_stability,
)
from portmodlib.atom import QualifiedAtom

from .globals import env
from .pybuild import Pybuild

if TYPE_CHECKING:
    from portmod.query import FlagDesc


class UseDep(NamedTuple):
    atom: QualifiedAtom
    flag: str
    description: Optional[FlagDesc]
    oldvalue: Optional[str]
    comment: Tuple[str, ...]

    def __repr__(self):
        if self.oldvalue:
            return f"UseDep({self.atom}, {self.oldvalue} -> {self.flag})"
        else:
            return f"UseDep({self.atom}, {self.flag})"


def is_keyword_masked(arch: str, keywords: Iterable[str]):
    return "-" + arch in keywords or (
        "-*" in keywords and arch not in keywords and "~" + arch not in keywords
    )


class KeywordDep(NamedTuple):
    """A requirement that a keyword be accepted before the package can be installed"""

    atom: QualifiedAtom
    keyword: Keyword
    masked: bool
    # Keyword which made this dependency necessary
    masking_keyword: Optional[Keyword] = None


class LicenseDep(NamedTuple):
    """A requirement that a license be accepted before the package can be installed"""

    atom: QualifiedAtom
    license: str
    is_eula: bool
    repo: str


def select_package(packages: Iterable[Pybuild]) -> Tuple[Pybuild, Any]:
    """
    Chooses a mod version based on keywords and accepts it if the license is accepted
    """
    if not packages:
        raise Exception("Cannot select mod from empty modlist")

    stable = []
    testing = []
    untested = []
    masked = []

    for package in packages:
        stability, masking_keyword = get_stability(package)
        if stability == Stability.STABLE:
            stable.append((package, masking_keyword))
        elif stability == Stability.TESTING:
            testing.append((package, masking_keyword))
        elif stability == Stability.MASKED:
            masked.append((package, masking_keyword))
        elif stability is Stability.UNTESTED:
            untested.append((package, masking_keyword))

    is_masked = False
    keyword: Optional[Keyword] = None
    masking_keyword = None

    if stable:
        package, _ = max(stable, key=lambda pkg: pkg[0].version)
    elif testing:
        # No packages were accepted. Choose the best version and add the keyword
        # as a requirement for it to be installed
        package, _ = max(testing, key=lambda pkg: pkg[0].version)
        keyword = NamedKeyword(env.prefix().ARCH, Stability.TESTING, None)
    elif untested:
        package, _ = max(untested, key=lambda pkg: pkg[0].version)
        keyword = WildcardKeyword.ALWAYS
    elif masked:
        package, masking_keyword = max(masked, key=lambda pkg: pkg[0].version)
        keyword = WildcardKeyword.ALWAYS
        is_masked = True

    deps: List[Any] = []
    if not is_license_accepted(package, package.get_use()):
        deps.append(
            LicenseDep(package.CPN, package.LICENSE, has_eula(package), package.REPO)
        )
    if keyword is not None:
        deps.append(
            KeywordDep(
                QualifiedAtom("=" + package.ATOM.CPF),
                keyword,
                is_masked,
                masking_keyword,
            )
        )

    return (package, deps or None)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for performing bulk queries on the mod database and repositories
"""

import logging
import multiprocessing
import os
import re
import shutil
import sys
from collections import defaultdict
from logging import info, warning
from multiprocessing import Process
from time import sleep
from typing import (
    AbstractSet,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from portmodlib.atom import Atom, FQAtom, QualifiedAtom, atom_sat
from portmodlib.colour import blue, bright, green, lblue, lgreen, red, yellow
from portmodlib.l10n import l10n
from portmodlib.portmod import Group, PackageIndexData, Person
from portmodlib.portmod import query as native_query
from portmodlib.portmod import update_index as native_update_index
from portmodlib.usestr import parse_usestr

from .config import get_config
from .config.mask import is_masked
from .config.use import get_use, get_use_expand, use_reduce
from .download import get_total_download_size
from .globals import env
from .loader import load_all, load_all_installed, load_installed_pkg, load_pkg
from .pybuild import InstalledPybuild, Pybuild
from .repo import LocalRepo, get_repo
from .repo.keywords import Stability, get_stability
from .repo.metadata import get_global_use
from .repo.metadata import get_package_metadata as get_native_metadata
from .repo.metadata import get_use_expand_values
from .typing import assert_never

Maintainer = Union[Group, Person]
Maintainers = Union[Maintainer, List[Maintainer]]


def get_maintainer_strings(maintainers: Maintainers) -> List[str]:
    if not isinstance(maintainers, list):
        maintainers = [maintainers]
    return [str(maintainer) for maintainer in maintainers]


def get_maintainer_string(maintainers: Maintainers) -> str:
    return list_maintainers_to_human_strings(get_maintainer_strings(maintainers))


def list_maintainers_to_human_strings(maintainers: List[str]) -> str:
    """return the list of maintainers as a human readible string"""
    result = ""
    for maintainer_id in range(len(maintainers)):
        maintainer = maintainers[maintainer_id]
        if maintainer_id >= len(maintainers) - 1:  # the last
            result += maintainer
        elif maintainer_id >= len(maintainers) - 2:  # the second last
            result += maintainer + " and "
        else:
            result += maintainer + ", "
    return result


def print_depgraph(
    mod: Pybuild, level: int, level_limit: int, seen: Set[Pybuild]
) -> int:
    """
    Recursively prints the dependency graph for the given mod.

    Note that use conditionals and or statements are ignored.
    This prints out all possible dependencies, not actual dependencies.
    """
    if level > level_limit:
        return level - 1

    deps = parse_usestr(mod.DEPEND + " " + mod.RDEPEND, token_class=Atom)

    if mod in seen and deps:
        print(" " * (level + 2) + "-- " + l10n("omit-already-displayed-tree"))
        return level - 1
    max_level = level

    seen.add(mod)

    def print_token(token, conditionals=None):
        atom = Atom(token)
        mod = max(load_pkg(atom), key=lambda pkg: pkg.version)
        enabled, _ = get_use(mod)

        def colour(flag):
            if flag.rstrip("?").lstrip("-!") in enabled:
                return bright(red(flag))
            return bright(blue(flag))

        use = map(colour, atom.USE)
        use_str = ""
        if atom.USE:
            use_str = f'[{" ".join(use)}]'

        if conditionals:
            dep = f"( {' '.join(conditionals)} ( {atom.strip_use()} ) ) "
        else:
            dep = f"({atom.strip_use()}) "

        print(" " * (level + 1) + f"-- {bright(green(mod.ATOM.CPF))} " + dep + use_str)
        return print_depgraph(mod, level + 1, level_limit, seen)

    def print_deps(deps, conditionals: List):
        nonlocal max_level
        if isinstance(deps, list):
            if not deps:
                return
            if isinstance(deps[0], str) and deps[0] == "||":
                for inner_token in deps[1:]:
                    print_deps(inner_token, conditionals)
            elif isinstance(deps[0], str) and deps[0].endswith("?"):
                for inner_token in deps[1:]:
                    print_deps(inner_token, conditionals + [deps[0]])
            else:
                for inner_token in deps:
                    print_deps(inner_token, conditionals)
        else:
            max_level = max(max_level, print_token(deps, conditionals))

    print_deps(deps, [])

    return max_level


def str_strip(value: str) -> str:
    return re.sub("(( +- +)|(:))", "", value)


def str_squelch_sep(value: str) -> str:
    return re.sub(r"[-_\s]+", " ", value)


def query(
    value: str,
    limit: int = 10,
) -> List[PackageIndexData]:
    """
    Finds mods that contain the given value in the given field
    """
    if not env.PREFIX_NAME:
        raise RuntimeError("Queries must be done inside a prefix!")
    return native_query(
        env.INDEX, [repo.name for repo in env.prefix().REPOS], value, limit
    )


# FIXME: For this to be indexed, we'll need to add DEPEND and RDEPEND to the index
# This also means indexing individual package versions as well as aggregated data
def query_depends(atom: Atom, all_mods=False) -> List[Tuple[FQAtom, str]]:
    """
    Finds mods that depend on the given atom
    """
    if all_mods:
        mods = load_all()
    else:
        mods = load_all_installed()

    depends = []
    for mod in mods:
        if not all_mods:
            enabled, disabled = get_use(mod)
            atoms = use_reduce(
                mod.RDEPEND + " " + mod.DEPEND,
                enabled,
                disabled,
                token_class=Atom,
                flat=True,
            )
        else:
            atoms = use_reduce(
                mod.RDEPEND + " " + mod.DEPEND,
                token_class=Atom,
                matchall=True,
                flat=True,
            )

        for dep_atom in atoms:
            if dep_atom != "||" and atom_sat(dep_atom, atom):
                depends.append((mod.ATOM, dep_atom))
    return depends


_T = TypeVar("_T", bound=Union[str, int])


def get_flag_string(
    name: Optional[_T],
    enabled: Iterable[_T],
    disabled: Iterable[_T],
    installed: Optional[AbstractSet[_T]] = None,
    *,
    verbose: bool = True,
    display_minuses=True,
):
    """
    Displays flag configuration

    Enabled flags are displayed as blue
    If the installed flag list is passed, flags that differ from the
    installed set will be green
    if name is None, the name prefix will be omitted and no quotes will
    surround the flags

    Also allows flags which are integers, to allow for proper sorting of
    TEXTURE_SIZE flags
    """

    def disable(value: _T) -> str:
        if display_minuses:
            return "-" + str(value)
        return str(value)

    flags = []
    for flag in sorted(enabled):
        if installed is not None and flag not in installed:
            flags.append(bright(lgreen(flag)))
        elif verbose:
            flags.append(red(bright(flag)))

    for flag in sorted(disabled):
        if installed is not None and flag in installed:
            flags.append(bright(lgreen(disable(flag))))
        elif verbose:
            if display_minuses:
                flags.append(blue(disable(flag)))
            else:
                flags.append(lblue(disable(flag)))

    inner = " ".join(flags)

    if not flags:
        return None

    if name:
        return f'{name}="{inner}"'

    return inner


def get_usestring(
    pkgs: Union[Pybuild, List[Pybuild]],
    installed_use: Optional[Set[str]] = None,
    enabled_use: Optional[Set[str]] = None,
    verbose: bool = True,
) -> str:
    """
    Displays flags for the given packages.

    Has two modes, single package, and multi-package, depending on whether a package,
    or a list of packages, was passed as the first argument, since some options only make
    sense in one of the modes and should be ignored otherwise.

    args:
        pkgs: Either a single package or a list of packages to display collectively
        installed_use: A list of flags to compare the enabled use flags to, highlighting changes.
            Should be set to None if not in single package mode
        enabled_use: Use flags which should be considered enabled for this package
            Should be set to None if not in single package mode.
        verbose: If True, always display all flags.
            Otherwise, only display flag changes (only relevant in single package mode).

    returns: A string representation of the flags and their states.
    """
    if enabled_use is None:
        enabled_use = set()

    if isinstance(pkgs, Pybuild):
        single_pkg = pkgs
        pkgs = [pkgs]
        matchall = False
    else:
        single_pkg = None
        matchall = True
        assert not installed_use, "installed_use cannot be set in multi-package mode"

    # Note: flags containing underscores are USE_EXPAND flags
    # and are displayed separately
    IUSE_STRIP = {
        flag for pkg in pkgs for flag in pkg.IUSE_EFFECTIVE if "_" not in flag
    }

    texture_options = {
        size
        for pkg in pkgs
        for size in use_reduce(
            pkg.TEXTURE_SIZES,
            enabled_use,
            matchall=matchall,
            flat=True,
            token_class=int,
        )
    }

    use_expand_strings = []
    for use in get_config().get("USE_EXPAND", []):
        if use in get_config().get("USE_EXPAND_HIDDEN", []):
            continue

        enabled_expand = set()
        disabled_expand = set()
        for pkg in pkgs:
            new_enabled, new_disabled = get_use_expand(
                pkg.IUSE_EFFECTIVE, enabled_use, use
            )
            enabled_expand |= new_enabled
            disabled_expand |= new_disabled

        if enabled_expand or disabled_expand:
            installed_expand: Optional[Set[str]] = None
            if installed_use is not None and single_pkg is not None:
                installed_expand, _ = get_use_expand(
                    single_pkg.IUSE_EFFECTIVE, installed_use, use
                )
            string = get_flag_string(
                use,
                enabled_expand,
                disabled_expand,
                installed_expand,
                verbose=verbose,
                display_minuses=not matchall,
            )
            use_expand_strings.append(string)

    if len(texture_options) >= 2:
        texture_size = next(
            iter(get_use_expand(enabled_use, enabled_use, "TEXTURE_SIZE")[0]),
            None,
        )
        if texture_size is not None or matchall:
            disabled_textures = (
                set(texture_options) - {int(texture_size)} if texture_size else set()
            )
            texture_string = get_flag_string(
                "TEXTURE_SIZE",
                [texture_size] if texture_size else [],
                sorted(disabled_textures if texture_size else texture_options),
                verbose=verbose,
                display_minuses=not matchall,
            )
        else:
            texture_string = ""
    else:
        texture_string = ""

    usestring = get_flag_string(
        "USE",
        enabled_use & IUSE_STRIP,
        IUSE_STRIP - enabled_use,
        installed_use,
        verbose=verbose,
        display_minuses=not matchall,
    )
    return " ".join(filter(None, [usestring] + use_expand_strings + [texture_string]))


class SearchResult:
    pkg: PackageIndexData

    def get_version_str(self, sortedmods: List[Pybuild], footnote) -> str:
        # List of version numbers, prefixed by either (~) or ** depending on
        # keyword for user's arch. Followed by use flags, including use expand

        # TODO: following eix
        # Details could be encoded in the Stability class
        # [M] for masked by profile
        # [m] for masked by user package.mask but not by profile
        # {M} for masked, but unmasked by user package.unmask
        # * masked by missing keyword but stable for some other arch
        # ** masked by missing keyword on all architectures (implemented, but overloaded
        # - explicitly keyword masked on a particular arch
        # ~ masked by ~keyword
        # (~) masked by ~keyword but overridden by package.accept_keywords
        # Note: they can combine
        # [M]~ masked by profile package.mask and masked by ~keyword
        # [m](~) masked by user package.mask and masked by ~keyword
        # TODO: What about displaying versions masked on?

        versions = []
        for mod in sortedmods:
            if mod.INSTALLED:
                continue

            stability, _ = get_stability(mod)
            if stability == Stability.STABLE:
                version_nofmt = mod.PVR
                version = green(mod.PVR)
            elif stability == Stability.TESTING:
                version_nofmt = "~" + mod.PVR
                version = yellow(version_nofmt)
            elif stability == Stability.MASKED:
                version_nofmt = "-" + mod.PVR
                version = red(version_nofmt)
            elif stability == Stability.UNTESTED:
                version_nofmt = "**" + mod.PVR
                version = red(version_nofmt)
            else:
                assert_never(stability)

            repo = mod.REPO

            if is_masked(mod.ATOM, mod.REPO):
                versions.append(red("[M]" + version_nofmt) + footnote(repo))
            else:
                versions.append(version + footnote(repo))
        return " ".join(versions)

    def __init__(self, pkg: PackageIndexData, footnote):
        self.pkg = pkg

        group = load_pkg(Atom(self.pkg.cpn))
        sortedmods = sorted(group, key=lambda pkg: pkg.version)
        newest = sortedmods[-1]
        installed = load_installed_pkg(Atom(self.pkg.cpn))
        download = "{:.3f} MiB".format(get_total_download_size([newest]) / 1024 / 1024)

        if installed is not None:
            installed_str = blue(bright(installed.PVR)) + footnote(installed.REPO)

            installed_flags = get_usestring(installed, None, installed.get_use())
            if installed_flags:
                installed_str += f" {{{installed_flags}}}"
        else:
            installed_str = "not installed"

        version_str = self.get_version_str(sortedmods, footnote)

        all_flags = get_usestring(sortedmods)
        if all_flags:
            version_str += f" {{{all_flags}}}"

        indent = "      "
        verbose = logging.root.level < logging.DEBUG
        result = ""
        if self.pkg.homepage:
            if verbose:
                formatted_homepages = f"{os.linesep}{indent}".join(
                    [self.pkg.homepage] + self.pkg.other_homepages
                )
            else:
                formatted_homepages = self.pkg.homepage

            homepage_str = (
                f"{indent}{green(l10n('package-homepage'))} {formatted_homepages}"
                + os.linesep
            )
        result += (os.linesep + indent).join(
            [
                f"{bright(green(self.pkg.name))} ({bright(self.pkg.cpn)})",
                f"{green(l10n('package-available-versions'))} {version_str}",
                f"{green(l10n('package-installed-version'))} {installed_str}",
            ]
        ) + os.linesep
        if verbose:
            result += (
                f"{indent}{green(l10n('package-size-of-files'))} {download}"
                + os.linesep
            )
        if self.pkg.homepage:
            result += homepage_str
        result += (
            f"{indent}{green(l10n('package-description'))} {str_squelch_sep(newest.DESC)}"
            + os.linesep
        )
        if verbose:
            result += (
                f"{indent}{green(l10n('package-license'))} {newest.LICENSE}"
                + os.linesep
            )
        if self.pkg.upstream_maintainers and verbose:
            result += (
                f"{indent}{green(l10n('package-upstream-author'))} {list_maintainers_to_human_strings(self.pkg.upstream_maintainers)}"
                + os.linesep
            )
        self.string = result

    def __str__(self):
        return self.string


class FlagDesc:
    """Use flag descriptions"""

    def __init__(self, desc: str):
        self.desc = desc

    def __str__(self):
        return self.desc


class LocalFlagDesc(FlagDesc):
    """Local use flag description"""

    def __init__(self, pkg: Pybuild, desc: str):
        super().__init__(desc)
        self.pkg = pkg

    def __repr__(self):
        return f"LocalDesc({self.pkg}, {self.desc})"


class UseExpandDesc(FlagDesc):
    """Local use flag description"""

    def __init__(self, category: str, flag: str, desc: str):
        super().__init__(desc)
        self.flag = flag
        self.category = category

    def __repr__(self):
        return f"UseExpandDesc({self.category}, {self.desc})"


def get_flag_desc(pkg: Pybuild, flag: str) -> Optional[FlagDesc]:
    """Returns the description for the given use flag"""
    repo_root = get_repo(pkg.REPO).location

    global_use = get_global_use(repo_root)
    metadata = get_native_metadata(pkg)

    if metadata and flag in metadata.use:
        return LocalFlagDesc(pkg, metadata.use[flag])
    if flag in global_use:
        return FlagDesc(global_use[flag])
    if flag.startswith("texture_size_"):
        size = flag.replace("texture_size_", "")
        return UseExpandDesc("texture_size", size, l10n("texture-size-desc", size=size))
    if "_" in flag:  # USE_EXPAND
        use_expand = flag.rsplit("_", 1)[0]
        suffix = flag.replace(use_expand + "_", "")
        use_expand_desc = get_use_expand_values(repo_root, use_expand).get(suffix)
        if use_expand_desc:
            return UseExpandDesc(use_expand, suffix, use_expand_desc)

    return None


def get_flags(
    pkg: Pybuild,
) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, Dict[str, str]]]:
    """
    Returns all use flags and their descriptions for the given package

    This will ignore flags if they are in USE_EXPAND_HIDDEN, so it should only be used
    for display purposes.

    returns:
        Three dictionaries, one each for local flags, global flags and use_expand flags,
        in that order. The use expand flags are subdivided for each use_expand category.
    """
    repo_root = get_repo(pkg.REPO).location

    global_use = get_global_use(repo_root)
    metadata = get_native_metadata(pkg)
    if not metadata and isinstance(pkg, InstalledPybuild):
        for remote_pkg in load_pkg(pkg.CPN):
            remote_metadata = get_native_metadata(remote_pkg)
            if remote_metadata:
                metadata = remote_metadata

    local_flags = {}
    global_flags = {}
    use_expand_flags: DefaultDict[str, Dict[str, str]] = defaultdict(dict)

    for flag in pkg.IUSE_EFFECTIVE:
        if metadata and flag in metadata.use:
            local_flags[flag] = metadata.use[flag]
        elif flag in global_use:
            global_flags[flag] = global_use[flag]
        elif flag.startswith("texture_size_"):
            size = flag.replace("texture_size_", "")
            desc = l10n("texture-size-desc", size=size)
            use_expand_flags["texture_size"][size] = desc
        elif "_" in flag:  # USE_EXPAND
            use_expand = flag.rsplit("_", 1)[0]
            if use_expand.upper() in get_config().get("USE_EXPAND_HIDDEN", set()):
                # Don't produce descriptions for hidden flags
                continue
            suffix = flag.replace(use_expand + "_", "")
            # Produce flag mapped to empty string if description cannot be found
            use_expand_desc = get_use_expand_values(repo_root, use_expand).get(
                suffix, ""
            )
            use_expand_flags[use_expand][suffix] = use_expand_desc
        else:
            # No description Found.
            # Might be an installed package without metadata.yaml
            local_flags[flag] = ""

    return local_flags, global_flags, use_expand_flags


def get_package_metadata(
    package: QualifiedAtom, repo: LocalRepo
) -> Optional[PackageIndexData]:
    packages = load_pkg(package, only_repo_root=repo.location)
    if not packages:
        return None

    pkgs = sorted(packages, key=lambda x: x.ATOM.version)
    newest = pkgs[-1]

    pkg_metadata = PackageIndexData(
        cpn=package.CPN,
        repo=repo.name,
        category=package.C,
        package=package.PN,
        name=newest.NAME,
        desc=newest.DESC,
    )

    homepages = use_reduce(newest.HOMEPAGE, matchall=True, flat=True)

    if homepages:
        pkg_metadata.homepage = homepages[0]
    if len(homepages) > 1:
        pkg_metadata.other_homepages = homepages[1:]

    if newest.LICENSE:
        pkg_metadata.license = newest.LICENSE

    for pkg in pkgs:
        metadata = get_native_metadata(pkg)
        if metadata:
            pkg_metadata.tags |= metadata.tags
            if metadata.longdescription:
                pkg_metadata.longdescription = metadata.longdescription
            if metadata.maintainer:
                pkg_metadata.maintainers = get_maintainer_strings(metadata.maintainer)
            if metadata.upstream:
                if metadata.upstream.maintainer:
                    pkg_metadata.upstream_maintainers = get_maintainer_strings(
                        metadata.upstream.maintainer
                    )
                if metadata.upstream.doc:
                    pkg_metadata.upstream_doc = metadata.upstream.doc
                if metadata.upstream.bugs_to:
                    pkg_metadata.upstream_bugs_to = metadata.upstream.bugs_to
                if metadata.upstream.changelog:
                    pkg_metadata.upstream_changelog = metadata.upstream.changelog

    return pkg_metadata


def _get_index_data(repo: LocalRepo) -> List[PackageIndexData]:
    def get_package_names():
        seen = set()
        for pkg in load_all(only_repo_root=repo.location):
            if pkg.CPN not in seen:
                seen.add(pkg.CPN)
                yield pkg.CPN

    info(f"Beginning index update for repo {repo.name}")
    if sys.stderr.isatty() or env.TESTING:
        try:
            from progressbar import GranularBar as Bar
        except ImportError:
            from progressbar import Bar  # type: ignore
        from progressbar import ETA, BouncingBar, Counter, ProgressBar, Timer, Variable

        package_names = set()
        bar = ProgressBar(
            redirect_stdout=True,
            widgets=[
                Variable("status", format="Collecting packages"),
                BouncingBar(),
                Counter(),
                " ",
                Timer(),
            ],
        )

        i = 0
        bar.start()
        for name in get_package_names():
            package_names.add(name)
            bar.update(i)
            i += 1
        bar.finish()

        package_data = []
        bar = ProgressBar(
            redirect_stdout=True,
            widgets=[
                Variable("status", format="Loading package data"),
                Bar(),
                Counter(),
                " ",
                ETA(),
            ],
            max_value=len(package_names),
        )
        bar.start()
        i = 0
        for cpn in package_names:
            metadata = get_package_metadata(cpn, repo)
            if metadata:
                package_data.append(metadata)
            bar.update(i)
            i += 1
        bar.finish()
    else:
        package_data = []
        for name in get_package_names():
            metadata = get_package_metadata(name, repo)
            if metadata:
                package_data.append(metadata)
    return package_data


def _run_progress_bar(pipe):
    from progressbar import AnimatedMarker, ProgressBar, UnknownLength, Variable

    bar = ProgressBar(
        widgets=[
            Variable("status", format="{variables.status}"),
            " ",
            AnimatedMarker(),
        ],
        max_value=UnknownLength,
        variables={"status": "Updating index"},
    )
    bar.start()
    while not pipe.poll():
        bar.update()
        sleep(0.1)
    bar.update(status="Done updating index.")
    bar.finish()


def _commit_index(repo: str, package_data: List[PackageIndexData]):
    os.makedirs(env.INDEX, exist_ok=True)

    if sys.stderr.isatty() or env.TESTING:
        pipe, child_pipe = multiprocessing.Pipe()
        process = Process(target=_run_progress_bar, args=(child_pipe,))
        process.start()
        try:
            native_update_index(env.INDEX, repo, package_data)
        except OSError as e:
            warning(f"{e}: Probably a schema change. Removing index and trying again.")
            # Try again but first remove old index. Probably a change in schema
            shutil.rmtree(env.INDEX)
            os.makedirs(env.INDEX)
            native_update_index(env.INDEX, repo, package_data)
        pipe.send(True)
        process.join()
    else:
        info("Updating index...")
        try:
            native_update_index(env.INDEX, repo, package_data)
        except OSError as e:
            warning(f"{e}: Probably a schema change. Removing index and trying again.")
            shutil.rmtree(env.INDEX)
            os.makedirs(env.INDEX)
            native_update_index(env.INDEX, repo, package_data)
        info("Done updating index.")


def update_index(repo: LocalRepo):
    package_data = []
    package_data.extend(_get_index_data(repo))
    _commit_index(repo.name, package_data)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmod.repo.keywords import NamedKeyword, Stability
from portmod.repo.keywords import WildcardKeyword as W
from portmod.repo.keywords import _accepts as accepts
from portmod.repo.keywords import _get_stability
from portmodlib.version import Version

from .env import setup_env, tear_down_env

N = NamedKeyword.from_string


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_stable_visibility():
    assert accepts([N("test")], [N("test")])
    assert not accepts([N("test")], [N("test2")])
    assert accepts([N("~test")], [N("test")])
    # See note: https://wiki.gentoo.org/wiki/KEYWORDS
    # ~* does not imply *
    assert not accepts([W.TESTING], [N("test")])
    assert accepts([W.STABLE], [N("test")])


def test_wild_stable_visibility():
    assert not accepts([W.STABLE], [])

    assert accepts([N("test"), W.STABLE], [N("test2")])
    assert accepts([N("~test"), W.STABLE], [N("test2")])
    assert accepts([W.STABLE], [N("test")])
    assert accepts([N("~test"), W.STABLE], [N("test")])


def test_wild_testing_visibility():
    assert not accepts([W.TESTING], [])
    # See note: https://wiki.gentoo.org/wiki/KEYWORDS
    # ~* does not imply *
    assert not accepts([W.TESTING], [N("test")])
    assert accepts([W.TESTING], [N("~test")])
    assert not accepts([N("~test"), W.TESTING], [N("test2")])
    assert accepts([N("~test"), W.TESTING], [N("~test2")])
    assert not accepts([N("test"), W.TESTING], [N("test2")])
    assert accepts([N("test"), W.TESTING], [N("~test2")])


def test_testing_visibility():
    assert not accepts([N("test")], [N("~test")])
    assert accepts([N("~test")], [N("~test")])
    assert accepts([W.TESTING], [N("~test")])
    assert not accepts([W.STABLE], [N("~test")])


def test_wild_visibility():
    assert accepts([W.ALWAYS], [N("test")])
    assert accepts([W.ALWAYS], [N("~test")])
    assert accepts([W.ALWAYS], [])


def test_untested_visibility():
    assert not accepts([N("test")], [])
    assert not accepts([N("~test")], [])
    assert accepts([W.ALWAYS], [])


def test_versions():
    assert accepts([N("test{==1.0}")], [N("test")])
    assert accepts([N("test{==1.0}")], [N("test{>=0.1}")])
    assert accepts([N("test{==1.0}")], [N("test{<2.1,>=1.0}")])

    assert accepts([N("test{>=1.0,<2.0}")], [N("test{<2.1,>=1.0}")])
    assert accepts([N("test{>=1.0,<2.0}")], [N("test{>=1.5}")])
    assert not accepts([N("test{>=1.0,<2.0}")], [N("test{>=2.0}")])


def test_valid_keyword_versions():
    """Tests that keyword versions allow external versions only"""
    N("test{==1.0.0_alpha2_pre1}")

    with pytest.raises(ValueError):
        N("test{==e1-1.0.0_alpha2_pre1}")
    with pytest.raises(ValueError):
        N("test{==1.0.0_alpha2_pre1-r2}")


def test_multiple_versions():
    assert accepts([N("test{==1.0}")], [N("test"), N("-test{<1.0}")])
    assert accepts([N("test{==1.0}")], [N("test{>=1.0}"), N("-test{<1.0}")])
    assert not accepts([N("test{==0.1}")], [N("test{>=1.0}"), N("-test{<1.0}")])


def test_version_overlap():
    """Tests that masked takes priority over stable, and stable takes priority over testing"""
    arch = "test"
    ver = Version("1.0")
    assert _get_stability(
        [N("test{==1.0}")], [N("-test{<1.0}"), N("test")], arch, ver
    ) == (Stability.STABLE, None)
    ver = Version("0.1")
    assert _get_stability(
        [N("test{==0.1}")], [N("-test{<1.0}"), N("test")], arch, ver
    ) == (Stability.MASKED, N("-test{<1.0}"))

    assert _get_stability(
        [N("test{==0.1}")], [N("~test{<1.0}"), N("test")], arch, ver
    ) == (Stability.STABLE, None)


def test_version_overlap_wildcards():
    """Tests that named keywords take priority over wildcards"""
    ver = Version("1.0")
    assert _get_stability([N("test{==1.0}")], [W("-*"), N("test")], "test", ver) == (
        Stability.STABLE,
        None,
    )
    ver = Version("0.1")
    assert _get_stability([N("test2{==0.1}")], [W("-*"), N("test")], "test2", ver) == (
        Stability.MASKED,
        W("-*"),
    )
    assert _get_stability([N("test{==0.1}")], [W("~*"), N("test")], "test", ver) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability(
        [N("test2{==0.1}")], [W("~*"), N("test"), N("~test")], "test2", ver
    ) == (Stability.TESTING, W("~*"))
    assert _get_stability(
        [N("test{==0.1}")], [N("~test"), W("*"), N("test2")], "test", ver
    ) == (Stability.STABLE, None)
    assert _get_stability(
        [N("test{==0.1}"), W("*")], [N("~test"), N("test2")], "test", ver
    ) == (Stability.STABLE, None)


def test_get_stability():
    arch = "test"
    assert _get_stability([N("test")], [N("test")], arch) == (Stability.STABLE, None)
    assert _get_stability([N("test")], [N("~test")], arch) == (
        Stability.TESTING,
        N("~test"),
    )
    assert _get_stability([N("test")], [], arch="test") == (Stability.UNTESTED, None)
    assert _get_stability([N("test")], [N("-test")], arch) == (
        Stability.MASKED,
        N("-test"),
    )

    assert _get_stability([N("test"), W.STABLE], [N("test2")], arch) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability([N("test"), W.TESTING], [N("~test2")], arch) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability([N("test")], [N("test2"), W.MASKED], arch) == (
        Stability.MASKED,
        W.MASKED,
    )


def test_get_stability_versioned():
    arch = "test"
    ver = Version("1.0.0")
    assert _get_stability([N("test{==1.0}")], [N("test")], arch, ver) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability([N("test{==1.0.0}")], [N("test{>=1.0}")], arch, ver) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability(
        [N("test{==0.1.0}")], [N("test{>=1.0}")], arch, Version("0.1.0")
    ) == (Stability.UNTESTED, None)
    # Masked keywords always take priority
    assert _get_stability(
        [N("test{==0.1.0}")],
        [N("test"), N("-test{==0.1.0}")],
        arch,
        Version("0.1.0"),
    ) == (Stability.MASKED, N("-test{==0.1.0}"))
    # Otherwise, as long as any keyword is stable, the package is stable
    # Even if a testing keyword also matches
    assert _get_stability(
        [N("test{==0.1.0}")], [N("~test"), N("test{<1.0}")], arch, Version("0.1.0")
    ) == (Stability.STABLE, None)


def test_get_stability_versioned_acceptall():
    """Tests that checking the stability of packages when accepting all versions is consistent

    This is used by package validation when checking the stability of dependencies to ensure
    that at least some version has the desired stability.
    """
    arch = "test"
    ver = Version("1.0.0")
    assert _get_stability([N("test")], [N("test{<1.0}")], arch, ver) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability([N("test")], [N("test{==0.1}")], arch, ver) == (
        Stability.STABLE,
        None,
    )
    assert _get_stability([N("test")], [N("~test{==0.0.1}")], arch, ver) == (
        Stability.TESTING,
        N("~test{==0.0.1}"),
    )
    assert _get_stability(
        [N("test")], [N("~test{==0.0.1}"), N("-test{0.1*}")], arch, ver
    ) == (Stability.TESTING, N("~test{==0.0.1}"))
    assert _get_stability(
        [N("test")], [N("~test{==0.0.1}"), N("-test")], arch, ver
    ) == (Stability.MASKED, N("-test"))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests some otherwise untested parts of the interface
"""

import io
import os
import sys

import pytest

from portmod._cli.main import main
from portmod._cli.merge import CLIMerge
from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmod.merge import merge
from portmodlib.atom import Atom

from .env import connected, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    data = setup_env("test")
    merge(["@world"], update=True, io=CLIMerge())
    yield data
    tear_down_env()


def test_validate():
    """Tests that validate works correctly"""
    sys.argv = ["portmod", "test", "merge", "test", "test2", "--no-confirm"]
    main()
    sys.argv = ["portmod", "test", "validate"]
    main()


@pytest.mark.skipif(not connected(), reason="Requires network access")
def test_sync():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "sync"]
    main()


@pytest.mark.parametrize("verbose", [True, False])
def test_info(verbose):
    """Tests that portmod info works correctly"""
    sys.argv = ["portmod", "test", "info"]
    if verbose:
        sys.argv.append("--verbose")
    with pytest.raises(SystemExit) as exc_info:
        main()
        assert exc_info.value.returncode == 0


def test_search(monkeypatch):
    """Tests that portmod search works correctly"""
    sys.argv = ["portmod", "test", "search", "test"]
    monkeypatch.setattr("sys.stdin", io.StringIO("\n"))
    main()


def test_use():
    """Tests that portmod use works correctly"""
    sys.argv = ["portmod", "test", "use", "-E", "foo"]
    main()
    sys.argv = ["portmod", "test", "merge", "test4", "--no-confirm"]
    main()
    pkg = load_installed_pkg(Atom("test/test4"))
    assert pkg
    assert "foo" in pkg.get_use()


def test_mirror():
    """Tests that `portmod mirror` works correctly"""
    sys.argv = ["portmod", "mirror", os.path.join(env.TMP_DIR, "mirror"), "test"]
    main()


def test_outdated(capsys):
    """Tests that `portmod <prefix> outdated` works correctly"""
    sys.argv = ["portmod", "test", "outdated"]
    main()
    captured = capsys.readouterr().out
    assert "All packages are up to date!\n" == captured
    merge(["test/test-1.0"], io=CLIMerge())
    # Clear output
    capsys.readouterr()
    sys.argv = ["portmod", "test", "outdated"]
    main()
    captured = capsys.readouterr().out
    print(captured)
    assert "All packages are up to date!" not in captured
    assert "1.0-r2" in captured
    assert "2.0" in captured
    assert "Ready for world update" in captured



import os

import docutils
from docutils.parsers.rst import Directive, Parser
from docutils.utils import new_document


class Summary(Directive):
    has_content = True

    def run(self):
        path = os.path.join(
            os.path.dirname(self.state.document.current_source), self.content[0]
        )

        self.state.document.settings.record_dependencies.add(path)

        with open(path) as file:
            document = new_document(path, self.state.document.settings)
            Parser().parse(file.read(), document)
            for paragraph in document.findall(docutils.nodes.paragraph):
                if paragraph.astext().strip():
                    if paragraph.astext().startswith("Duplicate implicit target name"):
                        continue
                    summary = paragraph
                    break

        return [summary]


def setup(app):
    app.add_directive("summary", Summary)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Importing values from this module fills them with information about the
pybuild file from which they were imported
"""
# Note: This module should be removed from sys.modules prior to importing to ensure that
# the cached version is not used instead.

import os
from pathlib import Path
from typing import Optional

from portmodlib.atom import VAtom

CATEGORY: str = ""
"""
The package's category.

E.g. ``base``
"""
P: str = ""
"""
The package name and version.

E.g.: ``example-suite-1.0``
"""
PF: str = ""
"""
The package name with version and revision.

E.g.: ``example-suite-1.0-r1``
"""
PN: str = ""
"""
The package name without version.

E.g.: ``example-suite``
"""
PV: str = ""
"""
The package's version without revision

E.g. ``1.0``
"""
PR: Optional[str] = None
"""
The package's revision

E.g. ``r1``
"""
PVR: str = ""
"""
The package's version and revision

E.g. ``1.0-r1``
"""


def _get_info(atom):
    return {
        "P": atom.P,
        "PF": atom.PF,
        "PN": atom.PN,
        "PV": atom.PV,
        "PR": atom.PR,
        "PVR": atom.PVR,
    }


def _set_info(filename):
    """Puts module information in the global scope so that it can be imported"""
    global CATEGORY
    if filename is not None:
        CATEGORY = Path(filename).resolve().parent.parent.name
        name, _ = os.path.splitext(os.path.basename(filename))
        atom = VAtom("{}/{}".format(CATEGORY, name))

        globals().update(_get_info(atom))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Interface for installing and removing packages"""

import argparse
import logging
import os
import sys
import traceback
from logging import error, info, warning
from shutil import move, rmtree
from typing import AbstractSet, Callable, Dict, Optional, Sequence, Tuple

from portmod._deps import DepError, PackageDoesNotExist
from portmod.download import get_filename
from portmod.globals import env
from portmod.io import (
    InstallIO,
    MergeIO,
    MergeMode,
    Message,
    PhaseFunction,
    Progress,
    RemoveIO,
    Task,
    Transaction,
)
from portmod.loader import AmbiguousAtom, SandboxedProcess, load_all
from portmod.lock import exclusive_lock
from portmod.merge import InteractiveError, deselect, global_updates, merge
from portmod.news import display_unread_message
from portmod.prompt import prompt_bool, prompt_options
from portmod.pybuild import Pybuild
from portmod.query import FlagDesc, LocalFlagDesc
from portmod.repo import get_repo
from portmod.repo.keywords import WildcardKeyword
from portmod.repo.metadata import get_license
from portmod.transactions import print_transactions
from portmod.util import KeywordDep, LicenseDep, UseDep
from portmodlib.atom import Atom, FQAtom, InvalidAtom
from portmodlib.colour import bright, green, lblue, lgreen, red, yellow
from portmodlib.fs import onerror
from portmodlib.l10n import l10n

from . import atom_metavar


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def filter_mods(mods):
    atoms = []
    os.makedirs(env.DOWNLOAD_DIR, exist_ok=True)

    for mod in mods:
        if os.path.isfile(mod):
            for atom in load_all():
                for source in atom.get_source_manifests(matchall=True):
                    if source.check_file(mod):
                        move(mod, get_filename(source.name))
                        atoms.append(atom.ATOM)
            if os.path.exists(mod):
                warning(l10n("no-package-for-file", file=os.path.basename(mod)))
        else:
            atoms.append(mod)

    return atoms


def add_merge_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "merge",
        help=l10n("merge-help"),
        description=l10n("merge-desc"),
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "packages",
        metavar=atom_metavar(archive=True, sets=True),
        help=l10n("package-help"),
        nargs="*",
    )
    parser.add_argument(
        "--ignore-default-opts",
        help=l10n("ignore-default-opts-help"),
        action="store_true",
    )
    parser.add_argument(
        "-c", "--depclean", help=l10n("depclean-help"), action="store_true"
    )
    parser.add_argument(
        "-x", "--auto-depclean", help=l10n("auto-depclean-help"), action="store_true"
    )
    parser.add_argument(
        "-C", "--unmerge", help=l10n("unmerge-help"), action="store_true"
    )
    parser.add_argument(
        "-1", "--oneshot", help=l10n("oneshot-help"), action="store_true"
    )
    parser.add_argument("-O", "--nodeps", help=l10n("nodeps-help"), action="store_true")
    parser.add_argument("-u", "--update", help=l10n("update-help"), action="store_true")
    parser.add_argument(
        "-n", "--noreplace", help=l10n("noreplace-help"), action="store_true"
    )
    parser.add_argument("-N", "--newuse", help=l10n("newuse-help"), action="store_true")
    parser.add_argument(
        "-e", "--emptytree", help=l10n("emptytree-help"), action="store_true"
    )
    parser.add_argument("-D", "--deep", help=l10n("deep-help"), action="store_true")
    parser.add_argument(
        "-w",
        "--select",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        metavar=l10n("yes-or-no"),
        help=l10n("merge-select-help"),
    )
    parser.add_argument(
        "--deselect",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        metavar=l10n("yes-or-no"),
        help=l10n("merge-deselect-help"),
    )
    parser.add_argument("--sort-vfs", help=argparse.SUPPRESS, action="store_true")
    parser.add_argument("--debug", help=l10n("merge-debug-help"), action="store_true")

    parser.set_defaults(func=merge_main)


@exclusive_lock()
def merge_main(args):
    atoms = filter_mods(args.packages)
    env.DEBUG = args.debug

    if args.nodeps and args.depclean:
        error(l10n("nodeps-depclean"))
        sys.exit(1)

    try:
        if atoms or args.depclean:
            # If deselect is supplied (is not None), only deselect if not removing.
            # If removing, remove normally, but deselect depending on supplied value.
            if args.deselect and not (args.unmerge or args.depclean):
                deselect(atoms)
            else:
                try:
                    merge(
                        atoms,
                        delete=args.unmerge,
                        depclean=args.depclean,
                        oneshot=args.oneshot,
                        verbose=args.verbose,
                        update=args.update or args.newuse or args.noreplace,
                        nodeps=args.nodeps,
                        deselect=args.deselect,
                        select=args.select,
                        auto_depclean=args.auto_depclean,
                        deep=args.deep,
                        emptytree=args.emptytree,
                        io=CLIMerge(),
                    )

                    # Note: When execeptions occur, TMP_DIR should be preserved
                    if not env.DEBUG and os.path.exists(env.TMP_DIR):
                        rmtree(env.TMP_DIR, onerror=onerror)
                        info(">>> " + l10n("cleaned-up", dir=env.TMP_DIR))
                except (
                    InvalidAtom,
                    PackageDoesNotExist,
                    AmbiguousAtom,
                    DepError,
                    InteractiveError,
                ) as e:
                    if args.debug:
                        traceback.print_exc()
                    error(f"{e}")
                    sys.exit(1)

        if args.sort_vfs:
            global_updates()

    finally:
        display_unread_message()


class CLIMerge(MergeIO):
    def display_transactions(
        self,
        mode: MergeMode,
        transactions: Sequence[Transaction],
        new_selected: AbstractSet[Pybuild],
    ):
        """
        Transaction list to display to the user

        args:
            transactions: List of transactions to perform, in order.
            new_selected: Packages which were selected for installation
                          these, along with previously selected packages
                          (use :py:func:`portmod.config.sets:is_selected`) should be
                          visually distinguished from packages which are installed
                          just as dependencies
        """
        print(l10n(mode.value))
        verbose = logging.root.level <= logging.DEBUG
        print_transactions(transactions, new_selected, verbose=verbose)
        print()

    def use_changes(self, use_flags: Sequence[UseDep], apply_callback: Task):
        """
        Display necessary flag changes to the user and prompt them to accept
        the changes

        args:
            use_flags: he use flag changes which are required
            apply_callback: Task, which when executed will apply the changes
        """
        display_flags: Dict[str, FlagDesc] = {}
        for use in use_flags:
            flag = use.flag.lstrip("-")
            desc = use.description
            if isinstance(desc, LocalFlagDesc):
                display_flags[Atom(desc.pkg.ATOM.CPF).use(flag)] = desc
            else:
                display_flags[flag] = desc or FlagDesc("<missing description>")

        for key, value in display_flags.items():
            print(l10n("use-flag-desc", desc=value, flag=bright(lgreen(key))))

        print()
        print(l10n("necessary-flag-changes"))
        for use in use_flags:
            comment = "\n    ".join(use.comment)
            if comment:
                comment += "\n    "
            if use.flag.startswith("-") and use.oldvalue == use.flag.lstrip("-"):
                print(
                    "   {}{} {} # {}".format(
                        comment,
                        lblue(use.atom),
                        red(use.flag),
                        l10n("enabled-comment"),
                    )
                )
            elif not use.flag.startswith("-") and use.oldvalue == "-" + use.flag:
                print(
                    "    {}{} {} # {}".format(
                        comment,
                        green(use.atom),
                        red(use.flag),
                        l10n("disabled-comment"),
                    )
                )
            else:
                print("    {}{} {}".format(comment, green(use.atom), red(use.flag)))
        if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
            apply_callback.run()
        else:
            sys.exit(1)

    def keyword_changes(self, keywords: Sequence[KeywordDep], apply_callback: Task):
        """
        Display necessary keyword changes to the user and prompt them to accept
        the changes

        args:
            keywords: The keyword changes which are required
            apply_callback: Task, which when executed will apply the changes
        """
        for keyword in keywords:
            if keyword.masked:
                error(
                    l10n(
                        "package-masked-keyword",
                        atom=keyword.atom.CPN,
                        keyword=keyword.masking_keyword,
                    )
                )
                sys.exit(1)

        print(l10n("necessary-keyword-changes"))
        for keyword in keywords:
            if keyword.keyword == WildcardKeyword.ALWAYS:
                c = red
            else:
                c = yellow
            print("    {} {}".format(green(keyword.atom), c(keyword.keyword)))

        if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
            apply_callback.run()
        else:
            sys.exit(1)

    def license_changes(self, licenses: Sequence[LicenseDep], apply_callback: Task):
        """
        Display licenses which need to be accepted and prompt the user to accept them

        args:
            licenses: The licenses which must be accepted
            apply_callback: Task, which when executed will apply the changes
        """
        warning(l10n("necessary-license-changes"))
        for license in licenses:
            print("    {} {}".format(green(license.atom), license.license))
            # For EULA licenses, display the license and prompt the user to accept
            if license.is_eula and env.INTERACTIVE:
                print()

                def indent(text: str) -> str:
                    return "\n".join(["    " + line for line in text.split("\n")])

                print("    ---", l10n("license-start", license=license.license), "---")
                print(
                    indent(
                        get_license(get_repo(license.repo).location, license.license)
                    )
                )
                print("    ---", l10n("license-end", license=license.license), "---")

            if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
                pass
            else:
                sys.exit(1)

        apply_callback.run()

    def pkg_nofetch(self, package: FQAtom, instructions: str):
        """
        Display fetch information for the package

        args:
            package: The package which contains files which could not be fethed
            instructions: The instructions produced by the package describing how
                         to manually fetch the files
        """
        print(bright(yellow(l10n("fetch-instructions", atom=package))))
        print(instructions)
        print()

    def rebuild_warning(self, packages: Sequence[Atom], message: str):
        """
        Display message about packages which need to be rebuilt

        args:
            packages: The package which need to be rebuilt
            message: Message to display
        """
        warning(message)
        for atom in packages:
            print(f"    {green(atom)}")
        print(
            l10n(
                "rebuild-prompt",
                command=lgreen(f"portmod {env.PREFIX_NAME} merge @rebuild"),
            )
        )

    def pkg_messages(self, package: FQAtom, messages: Sequence[Message]):
        """
        Display custom messages for the package

        atgs:
            package: The package which produced the custom messages
            messages: Messages to display
        """
        print()
        print(">>> " + l10n("pkg-messages", atom=bright(green(package))))
        for message in messages:
            if message.typ == Message.Type.WARNING:
                warning(message.text)
            elif message.typ == Message.Type.INFO:
                info(message.text)
        print()

    def pkg_pretend(self, package: FQAtom, message: str):
        """
        Display pkg_pretend messages for the package

        args:
            package: The package which produced the message
            message: The output from pkg_pretend
        """
        print(message)

    def space_warning(self, package: FQAtom, message: str):
        """
        Display message about insufficient space

        args:
            package: The package which contains files which could not be fetched
            message: Message to display
        """
        # FIXME: indicate which package caused the warning
        warning(message)

    def download_ready(self, start_callback: Task):
        """
        Called when all checks have passed and we are ready to download files

        args:
            start_callback: Task to indicate that the user wants to begin merge
        """
        if prompt_bool(l10n("begin-download-qn")):
            start_callback.run()
        else:
            sys.exit(1)

    def merge_ready(self, start_callback: Task):
        """
        Called when files have been downloaded and we are ready to merge
        User is expected to handle manual downloads before this point
        May be called again if manually downloaded files do not pass checks

        args:
            start_callback: Task to indicate that the user wants to begin merge
        """
        if prompt_bool(l10n("begin-merge-qn")):
            start_callback.run()
        else:
            sys.exit(1)

    def finished(self, message: str):
        """
        Called when the merge operation has finished. The message should be displayed to the user
        """

    def get_install_io(self, package: FQAtom) -> InstallIO:
        """
        Produce an InstallIO object for use when installing a package

        This will be called once for each package to be installed
        """
        return CLIInstall(package)

    def get_remove_io(self, package: FQAtom) -> RemoveIO:
        """
        Produce an InstallIO object for use when removing a package

        This will be called once for each package to be installed
        """
        return CLIRemove(package)

    def get_stepped_fetch_progress(
        self, step: int, max_steps: int
    ) -> Callable[[str, int, Optional[int]], Progress]:
        def _get_fetch_progress(
            filename: str, start: int, end: Optional[int]
        ) -> Progress:
            """Returns a progress for displaying remote file fetching progress"""
            return CLIFetchProgress(
                filename, start=start, end=end, step=(step, max_steps)
            )

        return _get_fetch_progress


class CLIProgress(Progress):
    def get_bar(self, start: int, end: Optional[int], step: Optional[Tuple[int, int]]):
        try:
            from progressbar import GranularBar as Bar
        except ImportError:
            from progressbar import Bar  # type: ignore
        from progressbar import (
            Percentage,
            ProgressBar,
            SimpleProgress,
            Timer,
            UnknownLength,
        )

        return ProgressBar(
            widgets=[
                Percentage(),
                " (",
                SimpleProgress(),
                ") ",
                Bar(),
                " ",
                Timer(),
            ],
            max_value=end or UnknownLength,
            initial_value=start,
            prefix=f"({step[0]}/{step[1]})" if step else "",
        )

    def __init__(
        self,
        *,
        start: int = 0,
        end: Optional[int] = None,
        step: Optional[Tuple[int, int]] = None,
    ):
        self.current = 0
        if sys.stdout.isatty():
            self.bar = self.get_bar(start=start, end=end, step=step)
            self.bar.start()
        else:
            from progressbar import NullBar

            self.bar = NullBar()

    def update(self, *, value: Optional[int] = None):
        if value:
            self.current = value
        else:
            self.current += 1
        self.bar.update(self.current)

    def done(self):
        self.bar.finish()


class CLIFetchProgress(CLIProgress):
    def __init__(
        self,
        _filename: str,
        start: int,
        end: Optional[int],
        step: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(start=start, end=end, step=step)

    def get_bar(self, start: int, end: Optional[int], step: Optional[Tuple[int, int]]):
        try:
            from progressbar import GranularBar as Bar
        except ImportError:
            from progressbar import Bar  # type: ignore
        from progressbar import (
            ETA,
            DataSize,
            FileTransferSpeed,
            Percentage,
            ProgressBar,
            UnknownLength,
        )

        return ProgressBar(
            widgets=[
                Percentage(),
                " ",
                Bar(),
                " ",
                ETA(),
                " ",
                FileTransferSpeed(),
                " ",
                DataSize(),
            ],
            max_value=end or UnknownLength,
            initial_value=start,
            prefix=f"({step[0]}/{step[1]})" if step else "",
        )


class CLIRemove(RemoveIO):
    def __init__(self, atom):
        super().__init__(atom, False)

    def begin_removal(self):
        pass

    def phase_function(self, function: PhaseFunction, process: SandboxedProcess):
        pass

    def finished_removal(self):
        pass

    def remove_files(self, count: int) -> Progress:
        """Should return something which can handle displaying file installation progress"""
        return CLIProgress(end=count)


class CLIInstall(CLIRemove, InstallIO):
    def __init__(self, atom):
        super().__init__(atom)
        self.overwrite_all = False

    def begin_install(self):
        pass

    def finished_install(self):
        pass

    def check_conflicts(self, count: int) -> Progress:
        """Should return something which can handle displaying file conflict progress"""
        return CLIProgress(end=count)

    def install_files(self, count: int) -> Progress:
        """Should return something which can handle displaying file installation progress"""
        return CLIProgress(end=count)

    def can_overwrite(self, path: str) -> bool:
        if env.INTERACTIVE and not self.overwrite_all:
            result = prompt_options(
                l10n("pkg-file-conflict-prompt", file=path),
                [
                    (l10n("yes-short"), l10n("overwrite")),
                    (l10n("no-short"), l10n("dont-overwrite")),
                    (l10n("always-short"), l10n("always-overwrite")),
                ],
            )
        else:
            result = l10n("yes-short")

        if result == l10n("always-short"):
            self.overwrite_all = True

        return result in (l10n("yes-short"), l10n("always-short")) or self.overwrite_all



from tempfile import mkdtemp

import pytest

from portmod.config import get_config
from portmod.prefix import add_prefix

from .env import setup_env

try:
    import PySide6  # noqa: F401

    from portmod._gui.config import Config
except ModuleNotFoundError as error:
    if error.name == "PySide6":
        pytest.skip("GUI tests when PySide6 is not installed", allow_module_level=True)
    raise error


def test_get_current_prefix():
    setup_env("test")

    # If GUI_SELECTED_PREFIX is unset, __init__ will set it to
    # the first prefix it finds.
    gui_config = Config()

    assert gui_config.get_current_prefix() == "test"


def test_get_prefixes():
    # Setup 3 test prefixes
    setup_env("test")
    temp2 = mkdtemp("portmod.test2")
    add_prefix("test2", "test", temp2, "test", ["test"])
    temp3 = mkdtemp("portmod.test3")
    add_prefix("test3", "test", temp3, "test", ["test"])

    gui_config = Config()

    prefixes = gui_config.get_prefixes()

    assert prefixes == ["test", "test2", "test3"]


class TestCheckConfig:
    def test_gui_selected_prefix_not_set(self):
        # Setup 3 test prefixes
        setup_env("test")
        temp2 = mkdtemp("portmod.test2")
        add_prefix("test2", "test", temp2, "test", ["test"])
        temp3 = mkdtemp("portmod.test3")
        add_prefix("test3", "test", temp3, "test", ["test"])

        # GUI_SELECTED_PREFIX should be None before portmod._gui.config.check_config() runs.
        assert not get_config().get("GUI_SELECTED_PREFIX")

        # Initializing Config object runs check_config().
        Config()

        # GUI_SELECTED_PREFIX should now be "test3", as it should be the first prefix in the
        # list of prefixes.
        assert get_config().get("GUI_SELECTED_PREFIX") == "test"

    def test_gui_selected_prefix_set(self):
        # Setup 3 test prefixes
        setup_env("test")
        temp2 = mkdtemp("portmod.test2")
        add_prefix("test2", "test", temp2, "test", ["test"])
        temp3 = mkdtemp("portmod.test3")
        add_prefix("test3", "test", temp3, "test", ["test"])

        gui_config = Config()

        assert get_config().get("GUI_SELECTED_PREFIX") == "test"

        gui_config.set_prefix("test2")

        assert get_config().get("GUI_SELECTED_PREFIX") == "test2"



from PySide6.QtCore import QObject, Slot

from portmodlib.l10n import l10n


class L10n(QObject):
    def __init__(self) -> None:
        super().__init__()

    @Slot(str, result=str)
    def tr(self, msg_id: str) -> str:
        return l10n(msg_id)



import os
import sys

import pytest

from portmod.globals import env


@pytest.mark.skipif(
    "CI" not in os.environ,
    reason="Paths may differ from system to system; this just tests the default paths",
)
def test_config_dir():
    """Test that the CONFIG_DIR matches what is in the documentation. This may"""
    if sys.platform == "linux":
        assert env.CONFIG_DIR == os.path.expanduser("~/.config/portmod")
    elif sys.platform == "win32":
        assert env.CONFIG_DIR == os.path.expanduser(
            r"~\AppData\Roaming\portmod\portmod\config"
        )
    elif sys.platform == "darwin":
        assert env.CONFIG_DIR == os.path.expanduser(
            "~/Library/Preferences/portmod.portmod"
        )



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import csv
import filecmp
import os
from collections import defaultdict
from difflib import unified_diff
from logging import debug, info, warning
from pathlib import Path
from string import Template
from typing import Dict, List, Optional

from portmodlib.execute import execute
from portmodlib.fs import get_hash, match
from portmodlib.l10n import l10n
from portmodlib.parsers.list import read_list

from .config import get_config_value, variable_data_dir
from .globals import env


def get_protected_path(file: str) -> str:
    new_file = file
    i = 1
    while os.path.exists(new_file):
        name, ext = os.path.splitext(new_file)
        if ext.startswith(".__cfg_protect"):
            new_file = name + f".__cfg_protect_{i}__"
            i += 1
        else:
            new_file = file + ".__cfg_protect__"
    return new_file


def is_protected(file: str) -> bool:
    protected = get_config_value("CFG_PROTECT")
    if isinstance(protected, list):
        for pattern in protected:
            if match(Path(file).absolute().relative_to(env.prefix().ROOT), pattern):
                return True
    elif isinstance(protected, str):
        if match(Path(file).absolute().relative_to(env.prefix().ROOT), protected):
            return True
    return False


def get_mergetool() -> Optional[str]:
    """Returns user-configured mergetool"""
    result = get_config_value("MERGE_TOOL")
    if result:
        return str(result)
    return None


def merge_files(source: str, dest: str):
    """Invokes user-configured mergetool"""
    mergetool = get_mergetool()
    assert mergetool
    execute(
        Template(mergetool).substitute(
            orig=f'"{source}"', new=f'"{dest}"', merged=f'"{dest}"'
        )
    )


def get_redirections() -> Dict[str, List[str]]:
    """
    Iterates over all previously made file redirections and returns the (non-empty)
    results
    """
    results = defaultdict(list)
    to_remove = []
    if os.path.exists(os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv")):
        with open(
            os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv"), "r"
        ) as file:
            reader = csv.reader(file)
            for index, row in enumerate(reader):
                if not row:
                    continue
                if len(row) != 2:
                    warning(f"Invalid redirection in {file} on line {index}: {row}")
                    continue
                dst = row[0]
                src = row[1]

                if os.path.exists(src) and os.stat(src).st_size != 0:
                    results[dst].append(src)
                else:
                    to_remove.append((src, dst))

    for src, dst in to_remove:
        remove_redirection(src, dst)

    filtered = defaultdict(list)
    blacklist_file = os.path.join(variable_data_dir(), "module-data", "file-blacklist")
    blacklist = set()
    if os.path.exists(blacklist_file):
        blacklist = set(read_list(blacklist_file))

    for dst, newfiles in results.items():
        for newfile in newfiles:
            # If file is identical, ignore
            if os.path.exists(dst) and filecmp.cmp(newfile, dst, shallow=False):
                remove_redirection(newfile, dst)
                os.remove(newfile)
                continue

            if dst in blacklist:
                info(l10n("skipped-blacklisted-file", file=dst))
                remove_redirection(newfile, dst)
                os.remove(newfile)
                continue
            filtered[dst].append(newfile)

    for dst, newfiles in filtered.items():
        hashes: Dict[str, str] = {}
        # Do it backwards both so that we can remove from the list without breaking the iterator,
        # and to ensure we keep the most recent version of any duplicates.
        for path in reversed(newfiles):
            hash_value = get_hash(path)[0]
            if hash_value in hashes:
                # Double check
                if filecmp.cmp(path, hashes[hash_value], shallow=False):
                    debug(f"Removing old duplicate update file {path}")
                    remove_redirection(path, dst)
                    newfiles.remove(path)
                    os.remove(path)
            else:
                hashes[hash_value] = path
    return filtered


def remove_redirection(src: str, dst: str):
    path = os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv")
    if os.path.exists(path):
        with open(path, "r") as file:
            reader = csv.reader(file)
            rows = list(reader)
        with open(path, "w") as output_file:
            writer = csv.writer(output_file)
            for row in rows:
                if not row:
                    continue
                row_dst = row[0]
                row_src = row[1]
                if row_dst != dst and row_src != src:
                    writer.writerow(row)


def get_changes(src: str, dst: str) -> List[str]:
    if os.path.islink(src):
        src_lines = [l10n("symlink-to", path=os.readlink(src)) + "\n"]
    else:
        try:
            with open(src, "r") as src_file:
                src_lines = src_file.readlines()
        except UnicodeDecodeError:
            src_lines = ["<" + l10n("binary-data") + ">\n"]
    dst_lines = []
    if os.path.lexists(dst):
        if os.path.islink(dst):
            dst_lines = [l10n("symlink-to", path=os.readlink(dst)) + "\n"]
        else:
            try:
                with open(dst, "r") as dst_file:
                    dst_lines = dst_file.readlines()
            except UnicodeDecodeError:
                dst_lines = ["<" + l10n("binary-data") + ">\n"]

    return list(unified_diff(dst_lines, src_lines, dst, src))



from tempfile import mkdtemp
from test.env import select_profile, setup_env
from test.merge import merge

import pytest
from pytest import fail, fixture

from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmod.prefix import add_prefix
from portmodlib.atom import Atom

try:
    import PySide6  # noqa: F401

    from portmod._gui.config import Config
    from portmod._gui.Manage.InstalledPackagesModel import InstalledPackagesProxyModel
    from portmod._gui.packages import get_installed_packages, get_local_flags
except ModuleNotFoundError as error:
    if error.name == "PySide6":
        pytest.skip("GUI tests when PySide6 is not installed", allow_module_level=True)
    raise error


@fixture(scope="module")
def single_model() -> InstalledPackagesProxyModel:
    setup_env("test")
    merge(["test/test"], nodeps=True)

    return InstalledPackagesProxyModel(get_installed_packages())


@fixture(scope="module")
def model() -> InstalledPackagesProxyModel:
    setup_env("test")
    merge(["test/test4"], nodeps=True)
    merge(["test/test"], nodeps=True)

    return InstalledPackagesProxyModel(get_installed_packages())


@fixture(scope="module")
def test4_model() -> InstalledPackagesProxyModel:
    setup_env("test")
    merge(["test/test4"], nodeps=True)

    return InstalledPackagesProxyModel(get_installed_packages())


def test_init(model):
    assert set(model.realModel._data) == set(get_installed_packages())
    assert set(model.realModel._data) == {
        load_installed_pkg(Atom("test/test")),
        load_installed_pkg(Atom("test/test4")),
    }


def test_get_atom(model):
    assert {model.getAtom(0), model.getAtom(1)} == {"test/test", "test/test4"}


def test_change_to_current_prefix(model):
    temp2 = mkdtemp("portmod.test2")
    add_prefix("test2", "test", temp2, "test", ["test"])
    select_profile("test")

    merge(["test/test7"], nodeps=True)

    # Adding a new prefix sets the current prefix to that one, so set it back.
    env.set_prefix("test")

    assert set(model.realModel._data) == {
        load_installed_pkg(Atom("test/test")),
        load_installed_pkg(Atom("test/test4")),
    }

    gui_config = Config()
    gui_config.set_prefix("test2")

    model.changeToCurrentPrefix()

    assert model.realModel._data == [
        load_installed_pkg(Atom("test/test7")),
    ]


# Note: Order matters.
# Due to side effects of setting up the models like installing packages,
# tests using a certain model need to be run in sequence


def test_get_name(single_model):
    assert single_model.get_name(0) == "Test"


def test_get_author(single_model):
    assert (
        single_model.get_author(0)
        == "someone <someone@example.org>, someone else, justamail@example.org, just a string"
    )


def test_get_version(single_model):
    assert single_model.get_version(0) == "2.0"


def test_get_size(single_model):
    assert single_model.get_size(0) == "0.0 B"


def test_get_license(single_model):
    assert single_model.get_license(0) == "test"


def test_get_description(single_model):
    assert single_model.get_description(0) == "Test Desc: foo-bar"


def test_get_homepage(single_model):
    assert single_model.get_homepage(0) == "https://example.org"


def test_get_local_flags_model(test4_model):
    installed_pybuild = load_installed_pkg(Atom("test/test4"))
    if installed_pybuild:
        flags = get_local_flags(installed_pybuild)
        flags_model = test4_model.get_local_flags_model(0)
        assert flags_model._data == flags
    else:
        fail()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Type hints for Native Rust Extension"""
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, TypeVar, Union

# Ideally this file could eventually be generated automatically
# See https://github.com/PyO3/pyo3/issues/510

_T = TypeVar("_T")

def multi_select_prompt(
    question: str, options: List[_T], help_message: Optional[str] = None
) -> List[_T]: ...

class Group:
    group: str

class Person:
    """
    An individual maintainer
    Either name or email is required.
    """

    # Maintainer's Name or Pseudonym
    name: Optional[str]
    # Maintainer's Email
    email: Optional[str]
    # Description. Can be used to describe the status of maintainership
    desc: Optional[str]

class GroupDeclaration:
    desc: str
    members: List[Person]

Maintainer = Union[Group, Person]
Maintainers = Union[Maintainer, List[Maintainer]]

class Upstream:
    # maintainers/authors of the original mod.
    maintainer: Optional[Maintainers]
    # URL where a changelog for the mod can be found. Must be version independent
    changelog: Optional[str]
    # URL where the location of the upstream documentation can be found.
    # The link must not point to any third party documentation and must be version independent
    doc: Optional[str]
    # A place where bugs can be reported in the form of an URL or an e-mail address prefixed with mailto:
    bugs_to: Optional[str]

class PackageMetadata:
    # Description of the package
    longdescription: Optional[str]
    # Maintainer, or list of maintainers for the package
    maintainer: Optional[Maintainers]
    # Use flags and their descriptions. Key is the flag name, value is the description
    use: Dict[str, str]
    # Tags for searching
    tags: Set[str]
    # Description of the package's upstream information.
    upstream: Optional[Upstream]

class CategoryMetadata:
    # Description of the category.
    longdescription: str
    # Maintainer, or list of maintainers for the category
    maintainer: Optional[Maintainers]

class News:
    # A short descriptive title
    title: str
    # Author's name and email address, in the form Real Name <email@address>
    author: str
    # Translator's name and email address, in the form Real Name <email@address>
    translator: Optional[str]
    # Date of posting, in yyyy-mm-dd format
    posted: str
    revision: str
    # Only supported format is 2.0
    news_item_format: str
    # Contents of the news article
    body: str
    # Required installed packages for the news to be displayed
    display_if_installed: Optional[str]
    # Required keywords for the news to be displayed
    display_if_keyword: Optional[str]
    # Required profiles for the news to be displayed
    display_if_profile: Optional[str]

def l10n_lookup(locale: str, msg_id: str, kwargs: Dict[str, Any]) -> str: ...
def get_masters(filename: str) -> List[str]: ...
def parse_yaml_dict(filename: str) -> Dict[str, str]: ...
def parse_yaml_dict_dict(filename: str) -> Dict[str, Dict[str, str]]: ...
def parse_news(filename: str) -> News: ...
def parse_groups(filename: str) -> Dict[str, GroupDeclaration]: ...
def parse_category_metadata(filename: str) -> CategoryMetadata: ...
def parse_package_metadata(filename: str) -> PackageMetadata: ...
def _get_hash(filename: str, algs: List[str], buffer_size: int) -> List[str]: ...

class PackageIndexData:
    """Data class for package metadata returned by query results"""

    cpn: str
    repo: str
    category: str
    package: str
    name: str
    desc: str
    homepage: Optional[str]
    other_homepages: List[str]
    license: Optional[str]
    longdescription: Optional[str]
    maintainers: List[str]
    upstream_maintainers: List[str]
    upstream_doc: Optional[str]
    upstream_bugs_to: Optional[str]
    upstream_changelog: Optional[str]
    tags: Set[str]

    def __init__(
        self, cpn: str, repo: str, category: str, package: str, name: str, desc: str
    ): ...

def update_index(index_path: str, repo: str, packages: List[PackageIndexData]): ...
def query(
    index_path: str, repos: List[str], query: str, limit: int
) -> List[PackageIndexData]: ...

class _Directories:
    cache_dir: Path
    download_dir: Optional[Path]
    config_dir: Path
    data_dir: Path

directories: _Directories

def parse_yaml_mapping(data: str, filename: str) -> Dict: ...



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
The module accessible within pybuilds

Note that this module should not be imported outside of pybuild files
"""

from portmodlib._deprecated import (  # noqa  # pylint: disable=unused-import
    File,
    InstallDir,
)
from portmodlib._deprecated.vfs import (  # noqa  # pylint: disable=unused-import
    find_file,
    list_dir,
)
from portmodlib.fs import patch_dir  # noqa  # pylint: disable=unused-import
from portmodlib.globals import download_dir as _download_dir
from portmodlib.masters import get_masters  # noqa  # pylint: disable=unused-import
from portmodlib.usestr import (  # noqa  # pylint: disable=unused-import
    check_required_use,
    use_reduce,
)

from ._pybuild import (  # noqa  # pylint: disable=unused-import
    Pybuild1,
    Pybuild2,
    apply_patch,
)

DOWNLOAD_DIR = _download_dir()

__all__ = [
    "version_gt",
    "patch_dir",
    "get_masters",
    "use_reduce",
    "find_file",
    "list_dir",
    "File",
    "InstallDir",
    "Pybuild1",
    "Pybuild2",
    "apply_patch",
    "DOWNLOAD_DIR",
]


# Duplicated here since the version in portmodlib.atom has been deprecated and will be removed
def version_gt(version1: str, version2: str) -> bool:
    """
    Version comparision function

    args:
        version1: A version string
        version2: Another version string

    returns:
        True if and only if version1 is greater than version2
    """
    from portmodlib.atom import Version

    return Version(version1) > Version(version2)



import sys
from typing import List

from PySide6 import QtCore

from portmod import prefix
from portmod.config import get_config, set_config_value
from portmod.globals import env


class Config(QtCore.QObject):
    """
    Manages prefixes. Handles intializing them, switching them, getting
    the current one, and checking if one is selected.
    """

    def __init__(self) -> None:
        super(Config, self).__init__()
        self.check_config()

    def check_config(self) -> None:
        """
        Checks to make sure everything is setup before launching the GUI.

        First, it checks to make sure there is at least one prefix, and prompts
        the user to create one if there isn't.

        Second, it set's the current prefix in the GUI to whatever the value of
        'GUI_SELECTED_PREFIX' is in the Portmod configuration file. If that value is
        unset, it sets it to the first prefix it finds in the list of prefixes.
        """
        # Checks if there are any prefixes
        if prefix.get_prefixes().keys():
            current_prefix = get_config().get("GUI_SELECTED_PREFIX")
            if current_prefix in prefix.get_prefixes().keys():
                self.set_prefix(current_prefix)
            # If GUI_SELECTED_PREFIX is unset, set it to the first prefix in get_prefixes()
            else:
                self.set_prefix(list(prefix.get_prefixes().keys())[0])
        else:
            # TODO: If there isn't at least one prefix, make the user create one
            print("NO PREFIXES DETECTED")
            sys.exit(-1)

    @QtCore.Slot(result=str)
    def get_current_prefix(self) -> str:
        """
        Retrieves the name of the prefix from 'GUI_SELECTED_PREFIX' in the Portmod
        configuration file.
        """
        # This will never be None, as check_config() makes sure it is set.
        current_prefix: str = get_config()["GUI_SELECTED_PREFIX"]
        return current_prefix

    @QtCore.Slot(result=list)
    def get_prefixes(self) -> List[str]:
        """
        Returns a list of each prefix name.
        """
        return list(prefix.get_prefixes().keys())

    @QtCore.Slot(str)
    def set_prefix(self, name: str) -> None:
        """
        Sets the internal prefix, and also sets the value of 'GUI_SELECTED_PREFIX'
        within the Portmod configuration file.
        """
        env.set_prefix(name)
        set_config_value("GUI_SELECTED_PREFIX", name, env.GLOBAL_PORTMOD_CONFIG)

    # TODO: Update based on changes in init consolidation (!568)
    # Also, create a setup wizard.

    # def add_prefix_repo(self, repo_name: str, pfx_name: str):
    #     enabled_repos = get_config()["REPOS"]
    #
    #     if repo_name and repo_name not in enabled_repos:
    #         enabled_repos.add(repo_name)
    #
    #     set_config_value("REPOS", " ".join(sorted(enabled_repos)))
    #     # Re-set prefix so that env.prefix().REPOS is updated
    #     env.set_prefix(pfx_name)
    #
    # def init_prefix(self, name, arch, profile, repos) -> None:
    #     prefix.add_prefix(name, arch)
    #     env.set_prefix(name)
    #     set_config_value("GUI_SELECTED_PREFIX", name, env.GLOBAL_PORTMOD_CONFIG)
    #     for repo in repos:
    #         self.add_prefix_repo(repo.name, name)
    #     set_profile(get_repo(arch).location + "/profiles/" + profile)



from typing import Dict, List, Tuple

from portmod.loader import load_all_installed
from portmod.pybuild import InstalledPybuild
from portmod.query import get_flags, query
from portmod.repo import LocalRepo
from portmod.repo.metadata import get_profiles
from portmodlib.portmod import PackageIndexData


class LocalFlag:
    def __init__(self, name: str, description: str, status: bool) -> None:
        self.name = name
        self.description = description
        self.status = status


def get_installed_packages() -> List[InstalledPybuild]:
    return list(load_all_installed())


def get_local_flags(pybuild: InstalledPybuild) -> Dict[str, Tuple[bool, str]]:
    """
    Returns a dictionary where the keys are USE flags, and the values are
    a tuple containing the flags status (whether it's enabled or not),
    and its description.
    """

    flags = get_flags(pybuild)[0]

    restructured_flags = {}
    for flag in flags:
        restructured_flags[flag] = (
            flag in pybuild.get_use(),  # Flag status
            flags[flag],  # Description
        )

    return restructured_flags


def search_packages(search_term: str) -> List[PackageIndexData]:
    return query(search_term, 0)


def list_profiles(arch: str, repos: List[LocalRepo]) -> List[str]:
    return [
        name for values in get_profiles(arch, repos).values() for _, name, _ in values
    ]



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import logging
import sys
from itertools import repeat
from multiprocessing import Pool
from typing import List, Sequence

from portmod._deps import DepError, resolve
from portmod.config.sets import is_selected
from portmod.loader import load_all_installed, load_pkg
from portmod.transactions import Transactions
from portmod.util import select_package
from portmodlib.atom import FQAtom
from portmodlib.colour import bright, green, lblue
from portmodlib.l10n import l10n


def tabulate(
    header: Sequence[str],
    rows: Sequence[Sequence[str]],
    lengths: Sequence[Sequence[int]],
):
    assert rows, "Rows must be nonempty!"
    print()
    cols = len(header)
    gap = 4
    # Add one space between each column
    maxlengths = [max(row[col] + gap for row in lengths) for col in range(cols)]

    for col in range(cols):
        print(header[col] + " " * (maxlengths[col] - len(header[col])), end="")
    print()
    # Dividing line is the width of each column, minus the gap since the width of each column
    # includes an extra gap at the end
    print("—" * (sum(maxlengths) - gap))

    for index, row in enumerate(rows):
        for col in range(cols):
            print(row[col] + " " * (maxlengths[col] - lengths[index][col]), end="")
        print()
    print()


def is_outdated(pkg, world_update: List[FQAtom], ignore_keywords: bool):
    pkgs = load_pkg(pkg.CPN)
    if not pkgs:
        # Package is installed but not available in the repos
        return (pkg, "None", l10n("outdated-not-available"))
    if ignore_keywords:
        newest = max(pkgs, key=lambda x: x.version)
    else:
        newest, keywords = select_package(pkgs)
    if newest.version > pkg.version:
        if any(atom == newest.ATOM for atom in world_update):
            updateable = l10n("outdated-world-update")
        else:
            try:
                transactions = resolve([newest.ATOM], [], {newest.ATOM}, set(), set())
                if any(
                    transaction.pkg.ATOM == newest.ATOM
                    for transaction in transactions.pkgs
                ):
                    updateable = l10n("outdated-manual-update")
                else:
                    updateable = l10n("outdated-manual-update-impossible")
            except DepError:
                updateable = l10n("unable-to-satisfy-dependencies")
        return (pkg, str(newest.version), updateable)
    return None


def outdated(args):
    """Displays outdated packages"""

    # Suppress logger to avoid dependency calculation messages
    logger = logging.getLogger()
    logger.disabled = True
    try:
        world_update = resolve(
            [], [], set(), set(), set("world"), deep=True, update=True
        )
    except DepError:
        world_update = Transactions()

    # Multiprocessing only works if we can fork (default on Linux),
    # as otherwise the other processes don't inherit the correct prefix information
    # Forking is unstable on macos and not supported on Windows
    # Unfortunately, this can be somewhat slow on macos/Windows as a result
    if sys.platform == "linux":
        with Pool() as pool:
            outdated = [
                x
                for x in pool.starmap(
                    is_outdated,
                    zip(
                        [
                            pkg
                            for pkg in load_all_installed()
                            if "local" not in pkg.PROPERTIES
                        ],
                        repeat(
                            [transaction.pkg.ATOM for transaction in world_update.pkgs]
                        ),
                        repeat(args.ignore_keywords),
                    ),
                )
                if x is not None
            ]
    else:
        outdated = []
        world_pkgs = [transaction.pkg.ATOM for transaction in world_update.pkgs]
        for pkg in load_all_installed():
            if "local" not in pkg.PROPERTIES:
                result = is_outdated(pkg, world_pkgs, args.ignore_keywords)
                if result is not None:
                    outdated.append(result)

    if not outdated:
        print(l10n("outdated-up-to-date"))
        return

    longest = 0
    for installed, _, _ in outdated:
        length = len(installed.PN)
        if length > longest:
            longest = length

    strings = []
    lengths = []
    for installed, newest, updateable in outdated:
        lengths.append(
            (len(installed.CPN), len(installed.PVR), len(newest), len(updateable))
        )
        if is_selected(installed.ATOM):
            strings.append(
                (
                    bright(green(installed.CPN)),
                    lblue(installed.version),
                    lblue(newest),
                    updateable,
                )
            )
        else:
            strings.append(
                (
                    green(installed.CPN),
                    lblue(installed.version),
                    lblue(newest),
                    updateable,
                )
            )
    tabulate(
        (
            l10n("outdated-heading-package"),
            l10n("outdated-heading-installed"),
            l10n("outdated-heading-available"),
            l10n("outdated-heading-update-status"),
        ),
        strings,
        lengths,
    )
    logger.disabled = False


def add_outdated_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "outdated", help=l10n("outdated-help"), parents=parents
    )
    parser.add_argument(
        "--ignore-keywords", action="store_true", help=l10n("outdated-ignore-keywords")
    )
    parser.set_defaults(func=outdated)



import logging
import sys

from portmodlib.l10n import l10n

try:
    import PySide6  # noqa: F401
except ModuleNotFoundError as error:
    if error.name == "PySide6":
        logging.error(l10n("pyside6-not-found-error"))
        sys.exit(-1)
    else:
        raise error


import signal
import sys
from pathlib import Path

from PySide6.QtGui import QGuiApplication
from PySide6.QtQml import QQmlApplicationEngine, qmlRegisterSingletonInstance

from portmod._gui.config import Config
from portmod._gui.l10n import L10n
from portmod._gui.Manage.InstalledPackagesModel import InstalledPackagesProxyModel
from portmod._gui.packages import get_installed_packages
from portmod._gui.Search.SearchPackagesModel import SearchedPackagesModel


def main() -> None:
    # Handle Ctrl+C
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Setup the application engine
    app = QGuiApplication(sys.argv)
    engine = QQmlApplicationEngine()

    # Pass Python objects to QML application
    config = Config()
    qmlRegisterSingletonInstance(Config, "io.portmod", 1, 0, "Config", config)  # type: ignore

    l10n = L10n()
    qmlRegisterSingletonInstance(Config, "io.portmod", 1, 0, "L10n", l10n)  # type: ignore

    installedPackagesModel = InstalledPackagesProxyModel(get_installed_packages())
    engine.rootContext().setContextProperty(
        "installedPackagesModel", installedPackagesModel
    )

    searchPackagesModel = SearchedPackagesModel()
    engine.rootContext().setContextProperty("searchPackagesModel", searchPackagesModel)

    # Load the QML file
    qml_file = Path(__file__).parent / "main.qml"
    engine.load(qml_file)

    # Start the app
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Dict

from portmod.globals import env
from portmod.lock import exclusive_lock
from portmod.merge import merge
from portmod.query import SearchResult, query
from portmod.repo import get_repo
from portmodlib.colour import lblue
from portmodlib.l10n import l10n
from portmodlib.portmod import multi_select_prompt

from .merge import CLIMerge


def search_main(args):
    pkgs = list(
        query(
            " ".join(args.query),
            limit=0,
        )
    )

    if not pkgs:
        print(l10n("packages-found", num=len(pkgs)))
        return

    footnotes: Dict[str, str] = {}
    footnotenum = "a"

    def footnote(repo: str):
        nonlocal footnotenum
        if repo not in footnotes:
            footnotes[repo] = str(footnotenum)
            footnotenum = chr(ord(footnotenum) + 1)
        return lblue(f"[{footnotes[repo]}]")

    formatted = [SearchResult(package, footnote) for package in pkgs]
    footer_string = f"Select packages to install with space, then install them with Enter.{os.linesep}Typing filters the list; Escape to quit"
    if footnotes:
        footer_string += os.linesep + os.linesep.join(
            f"{footnote(repo)} {repo} {get_repo(repo).location}" for repo in footnotes
        )

    if env.INTERACTIVE:
        try:
            selection = multi_select_prompt(
                "Packages to install:", formatted, footer_string
            )
        except (EOFError, KeyboardInterrupt):
            return
        if selection:
            with exclusive_lock():
                merge([result.pkg.cpn for result in selection], io=CLIMerge())
    else:
        for pkg in formatted:
            print(pkg)
        print(footer_string)
        print(l10n("packages-found", num=len(formatted)))


def add_search_parser(subparsers, parents):
    parser = subparsers.add_parser("search", help=l10n("search-help"), parents=parents)

    parser.add_argument(
        "query",
        nargs="+",
        metavar=l10n("query-placeholder"),
        help=l10n("search-query-help"),
    )
    parser.set_defaults(func=search_main)



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests file master detection
"""

import pytest

from portmod._cli.merge import CLIInstall
from portmod.loader import load_pkg
from portmod.package import install_pkg
from portmodlib.atom import Atom

from .env import connected, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    """sets up and tears down test environment"""
    yield setup_env("test")
    tear_down_env()


@pytest.mark.skipif(not connected(), reason="Requires network access")
def test_unpack():
    """
    Tests that src_unpack works as expected for Pybuild2

    Checks are inside the package
    """
    pkg = load_pkg(Atom("=test/quill-of-feyfolken-2.0.2-r1"))[0]
    install_pkg(pkg, set(), io=CLIInstall(pkg.ATOM))
